import warnings
from contextlib import nullcontext as does_not_raise
from copy import deepcopy

import jax
import jax.numpy as jnp
import numpy as np
import pynapple as nap
import pytest
from scipy.interpolate import splev

import nemos as nmo
from nemos import utils
from nemos.base_class import Base


@pytest.mark.parametrize(
    "arrays, expected_out",
    [
        ([jnp.zeros((10, 1)), np.zeros((10, 1))], jnp.zeros((10, 2))),
        ([np.zeros((10, 1)), np.zeros((10, 1))], jnp.zeros((10, 2))),
        (
            [np.zeros((10, 1)), nap.TsdFrame(t=np.arange(10), d=np.zeros((10, 1)))],
            nap.TsdFrame(t=np.arange(10), d=np.zeros((10, 2))),
        ),
        (
            [
                nap.TsdFrame(t=np.arange(10), d=np.zeros((10, 1))),
                nap.TsdFrame(t=np.arange(10), d=np.zeros((10, 1))),
            ],
            nap.TsdFrame(t=np.arange(10), d=np.zeros((10, 2))),
        ),
        (
            [
                nap.TsdTensor(t=np.arange(10), d=np.zeros((10, 1, 2))),
                nap.TsdTensor(t=np.arange(10), d=np.zeros((10, 1, 2))),
            ],
            nap.TsdTensor(t=np.arange(10), d=np.zeros((10, 2, 2))),
        ),
    ],
)
def test_concatenate_eval(arrays, expected_out):
    """Test various combination of arrays and pyapple time series."""
    out = utils.pynapple_concatenate_numpy(arrays, axis=1)
    if hasattr(expected_out, "times"):
        assert np.all(out.d == expected_out.d)
        assert np.all(out.t == expected_out.t)
        assert np.all(out.time_support.values == expected_out.time_support.values)
    else:
        assert np.all(out == expected_out)


@pytest.mark.parametrize(
    "axis, arrays, expected_shape",
    [
        (0, [jnp.zeros((10,)), np.zeros((10,))], 20),
        (0, [jnp.zeros((10, 1)), np.zeros((10, 1))], 20),
        (1, [jnp.zeros((10, 1)), np.zeros((10, 1))], 2),
        (
            2,
            [
                nap.TsdTensor(t=np.arange(10), d=np.zeros((10, 1, 2))),
                nap.TsdTensor(t=np.arange(10), d=np.zeros((10, 1, 2))),
            ],
            4,
        ),
    ],
)
def test_concatenate_axis(arrays, axis, expected_shape):
    """Test various combination of arrays and pyapple time series."""
    assert utils.pynapple_concatenate_jax(arrays, axis).shape[axis] == expected_shape
    assert utils.pynapple_concatenate_numpy(arrays, axis).shape[axis] == expected_shape


@pytest.mark.parametrize(
    "dtype, arrays",
    [
        (np.int32, [jnp.zeros((10, 1)), np.zeros((10, 1))]),
        (np.float32, [jnp.zeros((10, 1)), np.zeros((10, 1))]),
        (
            np.int32,
            [
                nap.TsdTensor(t=np.arange(10), d=np.zeros((10, 1, 2))),
                nap.TsdTensor(t=np.arange(10), d=np.zeros((10, 1, 2))),
            ],
        ),
    ],
)
def test_concatenate_type(arrays, dtype):
    """Test various combination of arrays and pyapple time series."""
    assert utils.pynapple_concatenate_jax(arrays, dtype=dtype, axis=1).dtype == dtype
    assert (
        utils.pynapple_concatenate_numpy(
            arrays, dtype=dtype, axis=1, casting="unsafe"
        ).dtype
        == dtype
    )


class TestPadding:

    @pytest.mark.parametrize(
        "predictor_causality", ["causal", "acausal", "anti-causal", ""]
    )
    @pytest.mark.parametrize("iterable", [[np.zeros([2, 4, 5]), np.zeros([1, 1, 10])]])
    def test_conv_type(self, iterable, predictor_causality):
        raise_exception = not (
            predictor_causality in ["causal", "anti-causal", "acausal"]
        )
        if raise_exception:
            with pytest.raises(ValueError, match="predictor_causality must be one of"):
                utils.nan_pad(iterable, 3, predictor_causality)
        else:
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore",
                    category=UserWarning,
                    message="With acausal filter, pad_size should probably be even",
                )
                utils.nan_pad(iterable, 3, predictor_causality)

    @pytest.mark.parametrize("iterable", [[np.zeros([2, 4, 5]), np.zeros([2, 4, 6])]])
    @pytest.mark.parametrize("pad_size", [0.1, -1, 0, 1, 2, 3, 5, 6])
    def test_padding_nan_causal(self, pad_size, iterable):
        raise_exception = (not isinstance(pad_size, int)) or (pad_size <= 0)
        if raise_exception:
            with pytest.raises(
                ValueError, match="pad_size must be a positive integer!"
            ):
                utils.nan_pad(iterable, pad_size, "anti-causal")
        else:
            padded = utils.nan_pad(iterable, pad_size, "causal")
            for trial in padded:
                print(trial.shape, pad_size)
            assert all(np.isnan(trial[:pad_size]).all() for trial in padded), (
                "Missing NaNs at the " "beginning of the array!"
            )
            assert all(not np.isnan(trial[pad_size:]).any() for trial in padded), (
                "Found NaNs at the " "end of the array!"
            )
            assert all(
                padded[k].shape[0] == iterable[k].shape[0] + pad_size
                for k in range(len(padded))
            ), "Size after padding doesn't match expectation. Should be T + window_size - 1."

    @pytest.mark.parametrize("iterable", [[np.zeros([2, 5, 4]), np.zeros([2, 6, 4])]])
    @pytest.mark.parametrize("pad_size", [0, 1, 2, 3, 5, 6])
    def test_padding_nan_anti_causal(self, pad_size, iterable):
        raise_exception = (not isinstance(pad_size, int)) or (pad_size <= 0)
        if raise_exception:
            with pytest.raises(
                ValueError, match="pad_size must be a positive integer!"
            ):
                utils.nan_pad(iterable, pad_size, "anti-causal")
        else:
            padded = utils.nan_pad(iterable, pad_size, "anti-causal")
            for trial in padded:
                print(trial.shape, pad_size)
            assert all(
                np.isnan(trial[trial.shape[0] - pad_size :]).all() for trial in padded
            ), ("Missing NaNs at the " "end of the array!")
            assert all(
                not np.isnan(trial[: trial.shape[0] - pad_size]).any()
                for trial in padded
            ), ("Found NaNs at the " "beginning of the array!")
            assert all(
                padded[k].shape[0] == iterable[k].shape[0] + pad_size
                for k in range(len(padded))
            ), "Size after padding doesn't match expectation. Should be T + window_size - 1."

    @pytest.mark.parametrize("iterable", [[np.zeros([2, 5, 4]), np.zeros([2, 6, 4])]])
    @pytest.mark.parametrize("pad_size", [-1, 0.2, 0, 1, 3, 5])
    def test_padding_nan_acausal(self, pad_size, iterable):
        raise_exception = (not isinstance(pad_size, int)) or (pad_size <= 0)
        if raise_exception:
            with pytest.raises(
                ValueError, match="pad_size must be a positive integer!"
            ):
                utils.nan_pad(iterable, pad_size, "acausal")

        else:
            init_nan, end_nan = pad_size // 2, pad_size - pad_size // 2
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore",
                    category=UserWarning,
                    message="With acausal filter, pad_size should probably be even",
                )
                padded = utils.nan_pad(iterable, pad_size, "acausal")
            for trial in padded:
                print(trial.shape, pad_size)
            assert all(np.isnan(trial[:init_nan]).all() for trial in padded), (
                "Missing NaNs at the " "beginning of the array!"
            )
            assert all(
                np.isnan(trial[trial.shape[0] - end_nan :]).all() for trial in padded
            ), ("Missing NaNs at the " "end of the array!")

            assert all(
                not np.isnan(trial[init_nan : trial.shape[0] - end_nan]).any()
                for trial in padded
            ), ("Found NaNs in " "the middle of the array!")
            assert all(
                padded[k].shape[0] == iterable[k].shape[0] + pad_size
                for k in range(len(padded))
            ), "Size after padding doesn't match expectation. Should be T + window_size - 1."

    @pytest.mark.parametrize(
        "dtype, expectation",
        [
            (
                np.int8,
                pytest.raises(
                    ValueError, match="conv_time_series must have a float dtype"
                ),
            ),
            (
                jax.numpy.int8,
                pytest.raises(
                    ValueError, match="conv_time_series must have a float dtype"
                ),
            ),
            (np.float32, does_not_raise()),
            (jax.numpy.float32, does_not_raise()),
        ],
    )
    def test_nan_pad_conv_dtype(self, dtype, expectation):
        iterable = np.arange(100).reshape(1, 100, 1, 1).astype(dtype)
        with expectation:
            utils.nan_pad(iterable, 10)

    @pytest.mark.parametrize("causality", ["causal", "acausal", "anti-causal"])
    @pytest.mark.parametrize("pad_size", [1, 2])
    @pytest.mark.parametrize(
        "array, axis, expectation",
        [
            (jnp.zeros((10,)), 0, does_not_raise()),
            (jnp.zeros((10, 11)), 0, does_not_raise()),
            (jnp.zeros((10, 11)), 1, does_not_raise()),
            (
                jnp.zeros((10,)),
                1,
                pytest.raises(
                    ValueError, match="'axis' must be smaller than the number "
                ),
            ),
            (
                jnp.zeros((10, 11)),
                2,
                pytest.raises(
                    ValueError, match="'axis' must be smaller than the number "
                ),
            ),
            (
                jnp.zeros((10, 11)),
                0.0,
                pytest.raises(
                    ValueError, match="`axis` must be a non negative integer"
                ),
            ),
            (
                jnp.zeros((10, 11)),
                "x",
                pytest.raises(
                    ValueError, match="`axis` must be a non negative integer"
                ),
            ),
        ],
    )
    def test_axis_compatibility(self, pad_size, array, causality, axis, expectation):
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                category=UserWarning,
                message="With acausal filter, pad_size should probably be even",
            )
            with expectation:
                utils.nan_pad(array, pad_size, causality, axis=axis)

    @pytest.mark.parametrize("causality", ["causal", "acausal", "anti-causal"])
    @pytest.mark.parametrize(
        "pad_size, expectation",
        [
            (
                -1,
                pytest.raises(ValueError, match="pad_size must be a positive integer"),
            ),
            (
                1.0,
                pytest.raises(ValueError, match="pad_size must be a positive integer"),
            ),
            (1, does_not_raise()),
            (2, does_not_raise()),
        ],
    )
    @pytest.mark.parametrize("array", [jnp.zeros((10,)), np.zeros((10, 11))])
    def test_pad_size_type(self, pad_size, array, causality, expectation):
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                category=UserWarning,
                message="With acausal filter, pad_size should probably be even",
            )
            with expectation:
                utils.nan_pad(array, pad_size, causality, axis=0)

    @pytest.mark.parametrize(
        "causality, pad_size, expectation",
        [
            ("causal", 1, does_not_raise()),
            (
                "acausal",
                1,
                pytest.warns(
                    UserWarning,
                    match="With acausal filter, pad_size should probably be even",
                ),
            ),
            ("anti-causal", 1, does_not_raise()),
            ("causal", 2, does_not_raise()),
            ("acausal", 2, does_not_raise()),
            ("anti-causal", 2, does_not_raise()),
        ],
    )
    def test_pad_window_size(self, pad_size, causality, expectation):
        array = jnp.zeros((10,))
        with expectation:
            utils.nan_pad(array, pad_size, causality, axis=0)


class TestShiftTimeSeries:

    @pytest.mark.parametrize(
        "predictor_causality, expectation",
        [
            ("causal", does_not_raise()),
            ("anti-causal", does_not_raise()),
            (
                "invalid",
                pytest.raises(ValueError, match="predictor_causality must be one of"),
            ),
        ],
    )
    def test_causality_validation(self, predictor_causality, expectation):
        """Ensure the function rejects invalid predictor_causality values."""
        time_series = np.array([1.0, 2.0, 3.0])
        with expectation:
            utils.shift_time_series(time_series, predictor_causality)

    @pytest.mark.parametrize(
        "dtype, expectation",
        [
            (np.float32, does_not_raise()),
            (
                np.int32,
                pytest.raises(ValueError, match="time_series must have a float dtype"),
            ),
        ],
    )
    def test_dtype_validation(self, dtype, expectation):
        """Check that the function raises an error for non-float data types."""
        time_series = np.array([1, 2, 3], dtype=dtype)
        with expectation:
            utils.shift_time_series(time_series, "causal")

    @pytest.mark.parametrize(
        "axis, expectation",
        [
            (0, does_not_raise()),  # Assuming time_series is 1D for simplicity
            (
                1,
                pytest.raises(
                    ValueError,
                    match="'axis' must be smaller than the number of dimensions",
                ),
            ),
        ],
    )
    def test_axis_validation(self, axis, expectation):
        """Validate the function's handling of the axis parameter."""
        time_series = np.array([1.0, 2.0, 3.0])
        with expectation:
            utils.shift_time_series(time_series, "causal", axis)

    @pytest.mark.parametrize(
        "shape, predictor_causality, axis, expectation",
        [
            ((5, 5), "causal", 1, does_not_raise()),
            ((5, 5), "anti-causal", 0, does_not_raise()),
            (
                (5, 5),
                "invalid",
                0,
                pytest.raises(ValueError, match="predictor_causality must be one of"),
            ),
        ],
    )
    def test_shift_in_multidimensional_array(
        self, shape, predictor_causality, axis, expectation
    ):
        """Ensure correct shifting in multidimensional arrays along a specified axis."""
        time_series = np.zeros(shape)
        with expectation:
            shifted_series = utils.shift_time_series(
                time_series, predictor_causality, axis
            )
            if expectation == does_not_raise():
                # Check for NaN at the expected location
                if predictor_causality == "causal":
                    assert np.isnan(
                        shifted_series.take(0, axis=axis)
                    ).all(), (
                        "First element along the axis should be NaN for causal shift."
                    )
                    # Ensure no NaNs elsewhere
                    assert not np.isnan(
                        shifted_series.take(
                            range(1, shifted_series.shape[axis]), axis=axis
                        )
                    ).any(), "Unexpected NaNs found in the array."
                else:  # anti-causal
                    assert np.isnan(
                        shifted_series.take(-1, axis=axis)
                    ).all(), "Last element along the axis should be NaN for anti-causal shift."
                    # Ensure no NaNs elsewhere
                    assert not np.isnan(
                        shifted_series.take(
                            range(0, shifted_series.shape[axis] - 1), axis=axis
                        )
                    ).any(), "Unexpected NaNs found in the array."

    @pytest.mark.parametrize(
        "predictor_causality, axis, expectation",
        [
            ("causal", 0, does_not_raise()),
            ("anti-causal", 0, does_not_raise()),
        ],
    )
    def test_shift_with_pytree(self, predictor_causality, axis, expectation):
        """Test shifting functionality with a pytree of arrays."""
        time_series = [np.array([1.0, 2.0, 3.0]), np.array([4.0, 5.0, 6.0])]
        with expectation:
            shifted_series = utils.shift_time_series(
                time_series, predictor_causality, axis
            )
            for original, shifted in zip(time_series, shifted_series):
                if predictor_causality == "causal":
                    expected = np.concatenate(([np.nan], original[:-1]))
                else:  # anti-causal
                    expected = np.concatenate((original[1:], [np.nan]))
                np.testing.assert_array_equal(shifted, expected)

    def test_causal_shift(self):
        # Assuming time_series shape is (1, 3, 2, 2)
        time_series = np.random.rand(1, 3, 2, 2).astype(np.float32)
        shifted_series = utils.shift_time_series(time_series, "causal", axis=1)

        assert np.isnan(
            shifted_series[0, 0]
        ).all(), "First time bin should be NaN for causal shift"
        assert np.array_equal(
            shifted_series[0, 1:], time_series[0, :-1]
        ), "Causal shift did not work as expected"

    def test_anti_causal_shift(self):
        time_series = np.random.rand(1, 3, 2, 2).astype(np.float32)
        shifted_series = utils.shift_time_series(time_series, "anti-causal", axis=1)

        assert np.isnan(
            shifted_series[0, -1]
        ).all(), "Last time bin should be NaN for anti-causal shift"
        assert np.array_equal(
            shifted_series[0, :-1], time_series[0, 1:]
        ), "Anti-causal shift did not work as expected"


# Sample functions to test
def correct_function(x):
    return jnp.sin(x)  # Returns a jax.numpy.ndarray and is differentiable


def non_ndarray_function(x):
    return [x, x]  # Not returning a jax.numpy.ndarray


def nondifferentiable_function(x):
    knots = np.linspace(0, 1, 10)
    tck = np.zeros(10)
    tck[0] = 1
    return splev(x, (knots, tck, 3), 0)  # Not differentiable


def non_preserving_shape_function(x):
    return jnp.sum(x)  # Does not preserve input shape, returns scalar


def scalar_function(x):
    return jnp.sum(x)  # Returns a scalar


# Test cases with the match parameter
def test_assert_returns_ndarray():
    # Should pass
    utils.assert_returns_ndarray(
        correct_function, [jnp.array([1.0])], "correct_function"
    )

    # Should fail and match the error message
    with pytest.raises(TypeError, match="must return a jax.numpy.ndarray"):
        utils.assert_returns_ndarray(
            non_ndarray_function, [jnp.array([1.0])], "non_ndarray_function"
        )


def test_assert_differentiable():
    # Should pass
    utils.assert_differentiable(correct_function, "correct_function")

    # Should fail and match the error message
    with pytest.raises(TypeError, match="is not differentiable"):
        utils.assert_differentiable(
            nondifferentiable_function, "nondifferentiable_function"
        )


def test_assert_preserve_shape():
    # Should pass
    utils.assert_preserve_shape(
        correct_function, [jnp.array([1.0, 2.0])], "correct_function", 0
    )

    # Should fail and match the error message
    with pytest.raises(ValueError, match="must preserve the input array shape"):
        utils.assert_preserve_shape(
            non_preserving_shape_function,
            [jnp.array([1.0, 2.0])],
            "non_preserving_shape_function",
            0,
        )


def test_assert_scalar_func():
    # Should pass
    utils.assert_scalar_func(
        scalar_function, [jnp.array([1.0, 2.0])], "scalar_function"
    )

    # Should fail and match the error message
    with pytest.raises(TypeError, match="should return a scalar"):
        utils.assert_scalar_func(
            correct_function, [jnp.array([1.0])], "correct_function"
        )


# Mock classes inheriting Base
class Example(Base):
    # break alphabetical order in init params
    def __init__(self, c, a, b=None, d=1):
        self.a = a
        self.b = b
        self.c = c
        self.d = d


class ComplexParam(Base):
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return f"ComplexParam({self.name})"


@pytest.mark.parametrize(
    "obj, exclude_keys, use_name_keys, expected",
    [
        # Test basic functionality
        (Example(a=1, b="text", c=None), None, [], "Example(a=1, b='text', d=1)"),
        # Exclude keys
        (Example(a=1, b="text", c=None), ["b"], [], "Example(a=1, d=1)"),
        # Use __name__ for specified keys
        (Example(a=1, b=str, c=None), None, ["b"], "Example(a=1, b=str, d=1)"),
        # Exclude and use __name__ together
        (Example(a=1, b=Base, c=None), ["a"], ["b"], "Example(b=Base, d=1)"),
        # No parameters in get_params
        (Base(), None, [], "Base()"),
        # Parameters with shape (e.g., numpy arrays)
        (Example(a=None, b=np.array([0]), c=None), None, [], "Example(d=1)"),
        # Complex object values
        (
            Example(a=ComplexParam("x"), b="value", c=None),
            None,
            [],
            "Example(a=ComplexParam(x), b='value', d=1)",
        ),
        # Falsey values excluded
        (Example(a=0, b=False, c=None), None, [], "Example(a=0, b=False, d=1)"),
        # Falsey values excluded2
        (Example(a=0, b=[], c={}), None, [], "Example(a=0, d=1)"),
        # function without the __name__
        (nmo.observation_models.PoissonObservations(deepcopy(jax.numpy.exp)),None, [], "PoissonObservations(inverse_link_function=<PjitFunction>)")
    ],
)
def test_format_repr(obj, exclude_keys, use_name_keys, expected):
    assert utils.format_repr(obj, exclude_keys, use_name_keys) == expected


def test_repr_multiline():
    # test with label
    bas = nmo.basis.MSplineEval(10, label="mylabel")
    assert (
        utils.format_repr(bas, multiline=True)
        == "'mylabel':\n    MSplineEval(\n        n_basis_funcs=10,\n        order=4\n    )"
    )
    # test without label
    bas = nmo.basis.MSplineEval(10)
    assert (
        utils.format_repr(bas, multiline=True)
        == "MSplineEval(\n    n_basis_funcs=10,\n    order=4\n)"
    )


def test_order_preservation():
    """Ensure parameter order matches the __init__ method."""
    obj = Example(a=1, b="text", c=1)
    assert utils.format_repr(obj) == "Example(c=1, a=1, b='text', d=1)"


def test_shape_param_exclusion():
    """Test exclusion of parameters with a shape attribute."""

    obj = Example(a=np.arange(3), b=1, c=None)
    assert utils.format_repr(obj) == "Example(b=1, d=1)"



"""Mixin classes for basis."""

from __future__ import annotations

import abc
import copy
import inspect
from functools import wraps
from itertools import chain
from typing import TYPE_CHECKING, Generator, Optional, Tuple, Union

import numpy as np
from numpy.typing import ArrayLike, NDArray
from pynapple import Tsd, TsdFrame, TsdTensor

from ..convolve import create_convolutional_predictor
from ..utils import _get_terminal_size
from ._transformer_basis import TransformerBasis

if TYPE_CHECKING:
    from ._basis import Basis


def set_input_shape_state(states: Tuple[str] = ("_input_shape_product",)):
    """
    Decorator to preserve input shape-related attributes during method execution.

    This decorator ensures that the attributes `_n_basis_input_` and `_input_shape_product`
    are copied from the original object (`self`) to the returned object (`klass`)
    after the wrapped method executes. It is intended to be used with methods that
    clone or create a new instance of the class, ensuring these critical attributes
    are retained for functionality such as cross-validation.

    Parameters
    ----------
    method :
        The method to be wrapped. This method is expected to return an object
        (`klass`) that requires the `_n_basis_input_` and `_input_shape_` attributes.
    attr_list

    Returns
    -------
    :
        The wrapped method that copies `_input_shape_product` and `_input_shape_` from
        the original object (`self`) to the new object (`klass`).

    Examples
    --------
    Applying the decorator to a method:

    >>> from functools import wraps
    >>> @set_input_shape_state
    ... def __sklearn_clone__(self):
    ...     klass = self.__class__(**self.get_params())
    ...     return klass

    The `_input_shape_product` and `_input_shape_` attributes of `self` will be
    copied to `klass` after the method executes.
    """

    def decorator(method):
        @wraps(method)
        def wrapper(self, *args, **kwargs):
            # Call the original method
            klass = method(self, *args, **kwargs)
            # Copy the specified attributes
            for attr_name in states:
                setattr(klass, attr_name, getattr(self, attr_name, None))
            return klass

        return wrapper

    return decorator


class AtomicBasisMixin:
    """Mixin class for atomic bases (i.e. non-composite)."""

    def __init__(self, n_basis_funcs: int):
        self._n_basis_funcs = n_basis_funcs
        self._input_shape_ = None
        self._check_n_basis_min()

    @set_input_shape_state(states=("_input_shape_product", "_input_shape_"))
    def __sklearn_clone__(self) -> Basis:
        """Clone the basis while preserving attributes related to input shapes.

        This method ensures that input shape attributes (e.g., `_input_shape_product`,
        `_input_shape_`) are preserved during cloning. Reinitializing the class
        as in the regular sklearn clone would drop these attributes, rendering
        cross-validation unusable.
        """
        klass = self.__class__(**self.get_params())
        return klass

    def _iterate_over_components(self) -> Generator:
        """Return a generator that iterates over all basis components.

        For atomic bases, the list is just [self].

        Returns
        -------
            A generator returning self, it will be chained in composite bases.

        """
        return (x for x in [self])

    def set_input_shape(self, xi: int | tuple[int, ...] | NDArray):
        """
        Set the expected input shape for the basis object.

        This method configures the shape of the input data that the basis object expects.
        ``xi`` can be specified as an integer, a tuple of integers, or derived
        from an array. The method also calculates the total number of input
        features and output features based on the number of basis functions.

        Parameters
        ----------
        xi :
            The input shape specification.
            - An integer: Represents the dimensionality of the input. A value of ``1`` is treated as scalar input.
            - A tuple: Represents the exact input shape excluding the first axis (sample axis).
              All elements must be integers.
            - An array: The shape is extracted, excluding the first axis (assumed to be the sample axis).

        Raises
        ------
        ValueError
            If a tuple is provided and it contains non-integer elements.

        Returns
        -------
        self :
            Returns the instance itself to allow method chaining.

        Notes
        -----
        All state attributes that depends on the input must be set in this method in order for
        the API of basis to work correctly. In particular, this method is called by ``setup_basis``,
        which is equivalent to ``fit`` for a transformer. If any input dependent state
        is not set in this method, then ``compute_features`` (equivalent to ``fit_transform``) will break.

        """
        if isinstance(xi, tuple):
            if not all(isinstance(i, int) for i in xi):
                raise ValueError(
                    f"The tuple provided contains non integer values. Tuple: {xi}."
                )
            shape = xi
        elif isinstance(xi, int):
            shape = () if xi == 1 else (xi,)
        else:
            shape = xi.shape[1:]

        n_inputs = (int(np.prod(shape)),)

        self._input_shape_ = [shape]

        # total number of input time series. Used  for slicing and reshaping
        self._input_shape_product = n_inputs
        return self

    def _check_input_shape_consistency(self, x: NDArray):
        """Check input consistency across calls."""
        # remove sample axis and squeeze
        shape = x.shape[1:]

        initialized = self._input_shape_ is not None
        is_shape_match = self._input_shape_[0] == shape
        if initialized and not is_shape_match:
            expected_shape_str = "(n_samples, " + f"{self._input_shape_[0]}"[1:]
            expected_shape_str = expected_shape_str.replace(",)", ")")
            raise ValueError(
                f"Input shape mismatch detected.\n\n"
                f"The basis `{self.__class__.__name__}` with label '{self.label}' expects inputs with "
                f"a consistent shape (excluding the sample axis). Specifically, the shape should be:\n"
                f"  Expected: {expected_shape_str}\n"
                f"  But got:  {x.shape}.\n\n"
                "Note: The number of samples (`n_samples`) can vary between calls of `compute_features`, "
                "but all other dimensions must remain the same. If you need to process inputs with a "
                "different shape, please create a new basis instance, or set a new input shape by calling "
                "`set_input_shape`."
            )

    @property
    def input_shape(self) -> NDArray:
        return self._input_shape_[0] if self._input_shape_ else None


class EvalBasisMixin:
    """Mixin class for evaluational basis."""

    def __init__(self, bounds: Optional[Tuple[float, float]] = None):
        self.bounds = bounds

    def _compute_features(self, *xi: ArrayLike | Tsd | TsdFrame | TsdTensor):
        """Evaluate basis at sample points.

        The basis is evaluated at the locations specified in the inputs. For example,
        ``compute_features(np.array([0, .5]))`` would return the array:

        .. code-block:: text

           b_1(0) ... b_n(0)
           b_1(.5) ... b_n(.5)

        where ``b_i`` is the i-th basis.

        Parameters
        ----------
        *xi:
            The input samples over which to apply the basis transformation. The samples can be passed
            as multiple arguments, each representing a different dimension for multivariate inputs.

        Returns
        -------
        :
            A matrix with the transformed features.

        """
        out = self._evaluate(*(np.reshape(x, (x.shape[0], -1)) for x in xi))
        return np.reshape(out, (out.shape[0], -1))

    def setup_basis(self, *xi: NDArray) -> Basis:
        """
        Set all basis states.

        This method corresponds sklearn transformer ``fit``. As fit, it must receive the input and
        it must set all basis states, i.e. ``kernel_`` and all the states relative to the input shape.
        The difference between this method and the transformer ``fit`` is in the expected input structure,
        where the transformer ``fit`` method requires the inputs to be concatenated in a 2D array, while here
        each input is provided as a separate time series for each basis element.

        Parameters
        ----------
        xi:
            Input arrays.

        Returns
        -------
        :
            The basis with ready for evaluation.
        """
        self.set_input_shape(*xi)
        return self

    def _set_input_independent_states(self) -> "EvalBasisMixin":
        """
        Compute all the basis states that do not depend on the input.

        For EvalBasisMixin, this method might not perform any operation but simply return the
        instance itself, as no kernel preparation is necessary.

        Returns
        -------
        self :
            The instance itself.

        """
        return self

    @property
    def bounds(self):
        """Range of values covered by the basis."""
        return self._bounds

    @bounds.setter
    def bounds(self, values: Union[None, Tuple[float, float]]):
        """Setter for bounds."""
        if values is not None and len(values) != 2:
            raise ValueError(
                f"The provided `bounds` must be of length two. Length {len(values)} provided instead!"
            )

        # convert to float and store
        try:
            self._bounds = values if values is None else tuple(map(float, values))
        except (ValueError, TypeError):
            raise TypeError("Could not convert `bounds` to float.")

        if values is not None and values[1] <= values[0]:
            raise ValueError(
                f"Invalid bound {values}. Lower bound is greater or equal than the upper bound."
            )


class ConvBasisMixin:
    """Mixin class for convolutional basis."""

    def __init__(self, window_size: int, conv_kwargs: Optional[dict] = None):
        self.kernel_ = None
        self.window_size = window_size
        self.conv_kwargs = {} if conv_kwargs is None else conv_kwargs

    def _compute_features(self, *xi: NDArray | Tsd | TsdFrame | TsdTensor):
        """Convolve basis functions with input time series.

        A bank of basis filters is convolved with the input data. All the dimensions
        except for the sample-axis are flattened, so that the method always returns a
        matrix.

        For example, if inputs are of shape (num_samples, 2, 3), the output will be
        ``(num_samples, num_basis_funcs * 2 * 3)``.

        Parameters
        ----------
        *xi:
            The input data over which to apply the basis transformation. The samples can be passed
            as multiple arguments, each representing a different dimension for multivariate inputs.

        Notes
        -----
        This method is intended to be 1-to-1 mappable to sklearn ``transform`` method of transformer. This
        means that for the method to be callable, all the state attributes have to be pre-computed in a
        method that is mappable to ``fit``, which for us is ``_fit_basis``. It is fundamental that both
        methods behaves like the corresponding transformer method, with the only difference being the input
        structure: a single (X, y) pair for the transformer, a number of time series for the Basis.

        """
        self._check_has_kernel()
        # before calling the convolve, check that the input matches
        # the expectation. We can check xi[0] only, since convolution
        # is applied at the end of the recursion on the 1D basis, ensuring len(xi) == 1.
        conv = create_convolutional_predictor(self.kernel_, *xi, **self._conv_kwargs)
        # make sure to return a matrix
        return np.reshape(conv, newshape=(conv.shape[0], -1))

    def setup_basis(self, *xi: NDArray) -> Basis:
        """
        Set all basis states.

        This method corresponds sklearn transformer ``fit``. As fit, it must receive the input and
        it must set all basis states, i.e. ``kernel_`` and all the states relative to the input shape.
        The difference between this method and the transformer ``fit`` is in the expected input structure,
        where the transformer ``fit`` method requires the inputs to be concatenated in a 2D array, while here
        each input is provided as a separate time series for each basis element.

        Parameters
        ----------
        xi:
            Input arrays.

        Returns
        -------
        :
            The basis with ready for evaluation.
        """
        self.set_kernel()
        self.set_input_shape(*xi)
        return self

    def _set_input_independent_states(self):
        """
        Compute all the basis states that do not depend on the input.

        For Conv mixin the only attribute is the kernel.
        """
        return self.set_kernel()

    def set_kernel(self) -> "ConvBasisMixin":
        """
        Prepare or compute the convolutional kernel for the basis functions.

        This method is called to prepare the basis functions for convolution operations
        in subclasses. It computes a
        kernel based on the basis functions that will be used for convolution with the
        input data. The specifics of kernel computation depend on the subclass implementation
        and the nature of the basis functions.

        Returns
        -------
        self :
            The instance itself, modified to include the computed kernel. This
            allows for method chaining and integration into transformation pipelines.

        Notes
        -----
        Subclasses implementing this method should detail the specifics of how the kernel is
        computed and how the input parameters are utilized.

        """
        self.kernel_ = self._evaluate(np.linspace(0, 1, self.window_size))
        return self

    @property
    def window_size(self):
        """Duration of the convolutional kernel in number of samples."""
        return self._window_size

    @window_size.setter
    def window_size(self, window_size):
        """Setter overwrite default setter the window size parameter."""
        self._check_window_size(window_size)
        self._window_size = window_size

    def _check_window_size(self, window_size):
        if window_size is None:
            raise ValueError("You must provide a window_size!")

        elif not (isinstance(window_size, int) and window_size > 0):
            raise ValueError(
                f"`window_size` must be a positive integer. {window_size} provided instead!"
            )

    @property
    def conv_kwargs(self):
        """The convolutional kwargs.

        Keyword arguments passed to :func:`nemos.convolve.create_convolutional_predictor`.
        """
        return self._conv_kwargs

    @conv_kwargs.setter
    def conv_kwargs(self, values: dict):
        """Check and set convolution kwargs."""
        self._check_convolution_kwargs(values)
        self._conv_kwargs = values

    @staticmethod
    def _check_convolution_kwargs(conv_kwargs: dict):
        """Check convolution kwargs settings.

        Raises
        ------
        ValueError:
            If ``axis`` is provided as an argument, and it is different from 0
            (samples must always be in the first axis).
        ValueError:
            If ``self._conv_kwargs`` include parameters not recognized or that do not have
            default values in ``create_convolutional_predictor``.
        """
        if "axis" in conv_kwargs:
            raise ValueError(
                "Setting the `axis` parameter is not allowed. Basis requires the "
                "convolution to be applied along the first axis (`axis=0`).\n"
                "Please transpose your input so that the desired axis for "
                "convolution is the first dimension (axis=0)."
            )
        convolve_params = inspect.signature(create_convolutional_predictor).parameters
        convolve_configs = {
            key
            for key, param in convolve_params.items()
            if param.default
            # prevent user from passing
            # `basis_matrix` or `time_series` in kwargs.
            is not inspect.Parameter.empty
        }
        if not set(conv_kwargs.keys()).issubset(convolve_configs):
            # do not encourage to set axis.
            convolve_configs = convolve_configs.difference({"axis"})
            # remove the parameter in case axis=0 was passed, since it is allowed.
            invalid = (
                set(conv_kwargs.keys())
                .difference(convolve_configs)
                .difference({"axis"})
            )
            raise ValueError(
                f"Unrecognized keyword arguments: {invalid}. "
                f"Allowed convolution keyword arguments are: {convolve_configs}."
            )

    def _check_has_kernel(self) -> None:
        """Check that the kernel is pre-computed."""
        if self.kernel_ is None:
            raise RuntimeError(
                "You must call `setup_basis` before `_compute_features` for Conv basis."
            )


class BasisTransformerMixin:
    """Mixin class for constructing a transformer."""

    def to_transformer(self) -> TransformerBasis:
        """
        Turn the Basis into a TransformerBasis for use with scikit-learn.

        Examples
        --------
        Jointly cross-validating basis and GLM parameters with scikit-learn.

        >>> import nemos as nmo
        >>> from sklearn.pipeline import Pipeline
        >>> from sklearn.model_selection import GridSearchCV
        >>> # load some data
        >>> X, y = np.random.normal(size=(30, 1)), np.random.poisson(size=30)
        >>> basis = nmo.basis.RaisedCosineLinearEval(10).set_input_shape(1).to_transformer()
        >>> glm = nmo.glm.GLM(regularizer="Ridge", regularizer_strength=1.)
        >>> pipeline = Pipeline([("basis", basis), ("glm", glm)])
        >>> param_grid = dict(
        ...     glm__regularizer_strength=(0.1, 0.01, 0.001, 1e-6),
        ...     basis__n_basis_funcs=(3, 5, 10, 20, 100),
        ... )
        >>> gridsearch = GridSearchCV(
        ...     pipeline,
        ...     param_grid=param_grid,
        ...     cv=5,
        ... )
        >>> gridsearch = gridsearch.fit(X, y)
        """
        return TransformerBasis(self)


class CompositeBasisMixin:
    """Mixin class for composite basis.

    Add overwrites concrete methods or defines abstract methods for composite basis
    (AdditiveBasis and MultiplicativeBasis).
    """

    def __init__(self, basis1: Basis, basis2: Basis):
        # deep copy to avoid changes directly to the 1d basis to be reflected
        # in the composite basis.
        self.basis1 = copy.deepcopy(basis1)
        self.basis2 = copy.deepcopy(basis2)

        # set parents
        self.basis1._parent = self
        self.basis2._parent = self

        # if all bases where set, then set input for composition.
        set_bases = [s is not None for s in self.input_shape]

        if all(set_bases):
            # pass down the input shapes
            self.set_input_shape(*self.input_shape)

    @property
    def input_shape(self):
        shapes = [
            *(bas1.input_shape for bas1 in self.basis1._iterate_over_components()),
            *(bas2.input_shape for bas2 in self.basis2._iterate_over_components()),
        ]
        return shapes

    @property
    def _input_shape_(self):
        return self.input_shape

    @property
    @abc.abstractmethod
    def n_basis_funcs(self):
        """Read only property for composite bases."""
        pass

    def setup_basis(self, *xi: NDArray) -> Basis:
        """
        Set all basis states.

        This method corresponds sklearn transformer ``fit``. As fit, it must receive the input and
        it must set all basis states, i.e. ``kernel_`` and all the states relative to the input shape.
        The difference between this method and the transformer ``fit`` is in the expected input structure,
        where the transformer ``fit`` method requires the inputs to be concatenated in a 2D array, while here
        each input is provided as a separate time series for each basis element.

        Parameters
        ----------
        xi:
            Input arrays.

        Returns
        -------
        :
            The basis with ready for evaluation.
        """
        # setup both input independent
        self._set_input_independent_states()

        # and input dependent states
        self.set_input_shape(*xi)

        return self

    def _set_input_independent_states(self):
        """
        Compute the input dependent states for traversing the composite basis.

        Returns
        -------
        :
            The basis with the states stored as attributes of each component.
        """
        self.basis1._set_input_independent_states()
        self.basis2._set_input_independent_states()

    def _check_input_shape_consistency(self, *xi: NDArray):
        """Check the input shape consistency for all basis elements."""
        self.basis1._check_input_shape_consistency(
            *xi[: self.basis1._n_input_dimensionality]
        )
        self.basis2._check_input_shape_consistency(
            *xi[self.basis1._n_input_dimensionality :]
        )

    def _iterate_over_components(self):
        """Return a generator that iterates over all basis components.

        Reimplements the default behavior by iteratively calling _iterate_over_components of the
        elements.

        Returns
        -------
            A generator looping on each individual input.
        """
        return chain(
            self.basis1._iterate_over_components(),
            self.basis2._iterate_over_components(),
        )

    @set_input_shape_state(states=("_input_shape_product",))
    def __sklearn_clone__(self) -> Basis:
        """Clone the basis while preserving attributes related to input shapes.

        This method ensures that input shape attributes (e.g., `_input_shape_product`,
        `_input_shape_`) are preserved during cloning. Reinitializing the class
        as in the regular sklearn clone would drop these attributes, rendering
        cross-validation unusable.
        The method also handles recursive cloning for composite basis structures.
        """
        # clone recursively
        basis1 = self.basis1.__sklearn_clone__()
        basis2 = self.basis2.__sklearn_clone__()
        klass = self.__class__(basis1, basis2)

        for attr_name in ["_input_shape_product"]:
            setattr(klass, attr_name, getattr(self, attr_name))
        return klass

    def set_input_shape(self, *xi: int | tuple[int, ...] | NDArray) -> Basis:
        """
        Set the expected input shape for the basis object.

        This method sets the input shape for each component basis in the basis.
        One ``xi`` must be provided for each basis component, specified as an integer,
        a tuple of integers, or an array. The method calculates and stores the total number of output features
        based on the number of basis functions in each component and the provided input shapes.

        Parameters
        ----------
        *xi :
            The input shape specifications. For every k,``xi[k]`` can be:
            - An integer: Represents the dimensionality of the input. A value of ``1`` is treated as scalar input.
            - A tuple: Represents the exact input shape excluding the first axis (sample axis).
              All elements must be integers.
            - An array: The shape is extracted, excluding the first axis (assumed to be the sample axis).

        Raises
        ------
        ValueError
            If a tuple is provided and it contains non-integer elements.

        Returns
        -------
        self :
            Returns the instance itself to allow method chaining.
        """
        self._input_shape_product = (
            *self.basis1.set_input_shape(
                *xi[: self.basis1._n_input_dimensionality]
            )._input_shape_product,
            *self.basis2.set_input_shape(
                *xi[self.basis1._n_input_dimensionality :]
            )._input_shape_product,
        )
        return self

    def __repr__(self, n=0):
        _, rows = _get_terminal_size()
        rows = rows // 4
        # number of nested composite bases
        n += 1
        tab = "    "
        try:
            basis1 = self.basis1.__repr__(n=n)
        except TypeError:
            basis1 = self.basis1
        try:
            basis2 = self.basis2.__repr__(n=n)
        except TypeError:
            basis2 = self.basis2
        if n < rows:
            rep = f"{self.__class__.__name__}(\n{n*tab}basis1={basis1},\n{n*tab}basis2={basis2},\n{(n-1)*tab})"
        elif n == rows:
            rep = f"{self.__class__.__name__}(\n{n*tab}...\n{(n-1)*tab})"
        else:
            rep = None
        return rep



"""Bases classes."""

# required to get ArrayLike to render correctly
from __future__ import annotations

from typing import Optional, Tuple

from numpy.typing import ArrayLike, NDArray

from ..typing import FeatureMatrix
from ._basis import add_docstring
from ._basis_mixin import AtomicBasisMixin, ConvBasisMixin, EvalBasisMixin
from ._decaying_exponential import OrthExponentialBasis
from ._identity import HistoryBasis, IdentityBasis
from ._raised_cosine_basis import RaisedCosineBasisLinear, RaisedCosineBasisLog
from ._spline_basis import BSplineBasis, CyclicBSplineBasis, MSplineBasis
from ._transformer_basis import TransformerBasis

__all__ = [
    "IdentityEval",
    "HistoryConv",
    "MSplineEval",
    "MSplineConv",
    "BSplineEval",
    "BSplineConv",
    "CyclicBSplineEval",
    "CyclicBSplineConv",
    "RaisedCosineLinearEval",
    "RaisedCosineLinearConv",
    "RaisedCosineLogEval",
    "RaisedCosineLogConv",
    "OrthExponentialEval",
    "OrthExponentialConv",
    "TransformerBasis",
]


def __dir__() -> list[str]:
    return __all__


class BSplineEval(EvalBasisMixin, BSplineBasis):
    """
    B-spline 1-dimensional basis functions.

    Implementation of the one-dimensional BSpline basis [1]_.

    Parameters
    ----------
    n_basis_funcs :
        Number of basis functions.
    order :
        Order of the splines used in basis functions. Must lie within ``[1, n_basis_funcs]``.
        The B-splines have (order-2) continuous derivatives at each interior knot.
        The higher this number, the smoother the basis representation will be.
    bounds :
        The bounds for the basis domain. The default ``bounds[0]`` and ``bounds[1]`` are the
        minimum and the maximum of the samples provided when evaluating the basis.
        If a sample is outside the bounds, the basis will return NaN.
    label :
        The label of the basis, intended to be descriptive of the task variable being processed.
        For example: velocity, position, spike_counts.


    References
    ----------
    .. [1] Prautzsch, H., Boehm, W., Paluszny, M. (2002). B-spline representation. In: Bézier and B-Spline
        Techniques. Mathematics and Visualization. Springer, Berlin, Heidelberg.
        https://doi.org/10.1007/978-3-662-04919-8_5

    Examples
    --------
    >>> from numpy import linspace
    >>> from nemos.basis import BSplineEval
    >>> n_basis_funcs = 5
    >>> order = 3
    >>> bspline_basis = BSplineEval(n_basis_funcs, order=order)
    >>> bspline_basis
    BSplineEval(n_basis_funcs=5, order=3)
    >>> sample_points = linspace(0, 1, 100)
    >>> basis_functions = bspline_basis.compute_features(sample_points)
    """

    def __init__(
        self,
        n_basis_funcs: int,
        order: int = 4,
        bounds: Optional[Tuple[float, float]] = None,
        label: Optional[str] = "BSplineEval",
    ):

        BSplineBasis.__init__(
            self,
            n_basis_funcs,
            mode="eval",
            order=order,
            label=label,
        )
        EvalBasisMixin.__init__(self, bounds=bounds)

    @add_docstring("split_by_feature", BSplineBasis)
    def split_by_feature(
        self,
        x: NDArray,
        axis: int = 1,
    ):
        r"""
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import BSplineEval
        >>> from nemos.glm import GLM
        >>> basis = BSplineEval(n_basis_funcs=6, label="one_input")
        >>> X = basis.compute_features(np.random.randn(20,))
        >>> split_features_multi = basis.split_by_feature(X, axis=1)
        >>> for feature, sub_dict in split_features_multi.items():
        ...        print(f"{feature}, shape {sub_dict.shape}")
        one_input, shape (20, 6)

        """
        return super().split_by_feature(x, axis=axis)

    @add_docstring("_compute_features", EvalBasisMixin)
    def compute_features(self, xi: ArrayLike) -> FeatureMatrix:
        """
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import BSplineEval

        >>> # Generate data
        >>> num_samples = 1000
        >>> X = np.random.normal(size=(num_samples, ))  # raw time series
        >>> basis = BSplineEval(10)
        >>> features = basis.compute_features(X)  # basis transformed time series
        >>> features.shape
        (1000, 10)

        """
        return super().compute_features(xi)

    @add_docstring("evaluate_on_grid", BSplineBasis)
    def evaluate_on_grid(self, n_samples: int) -> Tuple[NDArray, NDArray]:
        """
        Examples
        --------
        Evaluate and visualize 4 B-spline basis functions of order 3:

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from nemos.basis import BSplineEval
        >>> bspline_basis = BSplineEval(n_basis_funcs=4, order=3)
        >>> sample_points, basis_values = bspline_basis.evaluate_on_grid(100)
        >>> for i in range(4):
        ...     p = plt.plot(sample_points, basis_values[:, i], label=f'Function {i+1}')
        >>> plt.title('B-Spline Basis Functions')
        Text(0.5, 1.0, 'B-Spline Basis Functions')
        >>> plt.xlabel('Domain')
        Text(0.5, 0, 'Domain')
        >>> plt.ylabel('Basis Function Value')
        Text(0, 0.5, 'Basis Function Value')
        >>> l = plt.legend()
        """
        return super().evaluate_on_grid(n_samples)

    @add_docstring("set_input_shape", AtomicBasisMixin)
    def set_input_shape(self, xi: int | tuple[int, ...] | NDArray):
        """
        Examples
        --------
        >>> import nemos as nmo
        >>> import numpy as np
        >>> basis = nmo.basis.BSplineEval(5)
        >>> # Configure with an integer input:
        >>> _ = basis.set_input_shape(3)
        >>> basis.n_output_features
        15
        >>> # Configure with a tuple:
        >>> _ = basis.set_input_shape((4, 5))
        >>> basis.n_output_features
        100
        >>> # Configure with an array:
        >>> x = np.ones((10, 4, 5))
        >>> _ = basis.set_input_shape(x)
        >>> basis.n_output_features
        100

        """
        return AtomicBasisMixin.set_input_shape(self, xi)


class BSplineConv(ConvBasisMixin, BSplineBasis):
    """
    B-spline 1-dimensional basis functions.

    Implementation of the one-dimensional BSpline basis [1]_.

    Parameters
    ----------
    n_basis_funcs :
        Number of basis functions.
    window_size :
        The window size for convolution in number of samples.
    order :
        Order of the splines used in basis functions. Must lie within ``[1, n_basis_funcs]``.
        The B-splines have (order-2) continuous derivatives at each interior knot.
        The higher this number, the smoother the basis representation will be.
    label :
        The label of the basis, intended to be descriptive of the task variable being processed.
        For example: velocity, position, spike_counts.
    conv_kwargs:
        Additional keyword arguments passed to :func:`nemos.convolve.create_convolutional_predictor`;
        These arguments are used to change the default behavior of the convolution.
        For example, changing the ``predictor_causality``, which by default is set to ``"causal"``.
        Note that one cannot change the default value for the ``axis`` parameter. Basis assumes
        that the convolution axis is ``axis=0``.

    References
    ----------
    .. [1] Prautzsch, H., Boehm, W., Paluszny, M. (2002). B-spline representation. In:
        Bézier and B-Spline Techniques. Mathematics and Visualization. Springer, Berlin, Heidelberg.
        https://doi.org/10.1007/978-3-662-04919-8_5

    Examples
    --------
    >>> from numpy import linspace
    >>> from nemos.basis import BSplineConv
    >>> n_basis_funcs = 5
    >>> order = 3
    >>> bspline_basis = BSplineConv(n_basis_funcs, order=order, window_size=10)
    >>> bspline_basis
    BSplineConv(n_basis_funcs=5, window_size=10, order=3)
    >>> sample_points = linspace(0, 1, 100)
    >>> features = bspline_basis.compute_features(sample_points)
    """

    def __init__(
        self,
        n_basis_funcs: int,
        window_size: int,
        order: int = 4,
        label: Optional[str] = "BSplineConv",
        conv_kwargs: Optional[dict] = None,
    ):
        ConvBasisMixin.__init__(self, window_size=window_size, conv_kwargs=conv_kwargs)
        BSplineBasis.__init__(
            self,
            n_basis_funcs,
            mode="conv",
            order=order,
            label=label,
        )

    @add_docstring("split_by_feature", BSplineBasis)
    def split_by_feature(
        self,
        x: NDArray,
        axis: int = 1,
    ):
        r"""
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import BSplineConv
        >>> from nemos.glm import GLM
        >>> basis = BSplineConv(n_basis_funcs=6, window_size=10, label="two_inputs")
        >>> X_multi = basis.compute_features(np.random.randn(20, 2))
        >>> split_features_multi = basis.split_by_feature(X_multi, axis=1)
        >>> for feature, sub_dict in split_features_multi.items():
        ...        print(f"{feature}, shape {sub_dict.shape}")
        two_inputs, shape (20, 2, 6)

        """
        return super().split_by_feature(x, axis=axis)

    @add_docstring("_compute_features", ConvBasisMixin)
    def compute_features(self, xi: ArrayLike) -> FeatureMatrix:
        """
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import BSplineConv

        >>> # Generate data
        >>> num_samples = 1000
        >>> X = np.random.normal(size=(num_samples, ))  # raw time series
        >>> basis = BSplineConv(10, window_size=11)
        >>> features = basis.compute_features(X)  # basis transformed time series
        >>> features.shape
        (1000, 10)

        """
        return super().compute_features(xi)

    @add_docstring("evaluate_on_grid", BSplineBasis)
    def evaluate_on_grid(self, n_samples: int) -> Tuple[NDArray, NDArray]:
        """
        Examples
        --------
        Evaluate and visualize 4 B-spline basis functions of order 3:

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from nemos.basis import BSplineConv
        >>> bspline_basis = BSplineConv(n_basis_funcs=4, order=3, window_size=10)
        >>> sample_points, basis_values = bspline_basis.evaluate_on_grid(100)
        >>> for i in range(4):
        ...     p = plt.plot(sample_points, basis_values[:, i], label=f'Function {i+1}')
        >>> plt.title('B-Spline Basis Functions')
        Text(0.5, 1.0, 'B-Spline Basis Functions')
        >>> plt.xlabel('Domain')
        Text(0.5, 0, 'Domain')
        >>> plt.ylabel('Basis Function Value')
        Text(0, 0.5, 'Basis Function Value')
        >>> l = plt.legend()
        """
        return super().evaluate_on_grid(n_samples)

    @add_docstring("set_input_shape", AtomicBasisMixin)
    def set_input_shape(self, xi: int | tuple[int, ...] | NDArray):
        """
        Examples
        --------
        >>> import nemos as nmo
        >>> import numpy as np
        >>> basis = nmo.basis.BSplineConv(5, 10)
        >>> # Configure with an integer input:
        >>> _ = basis.set_input_shape(3)
        >>> basis.n_output_features
        15
        >>> # Configure with a tuple:
        >>> _ = basis.set_input_shape((4, 5))
        >>> basis.n_output_features
        100
        >>> # Configure with an array:
        >>> x = np.ones((10, 4, 5))
        >>> _ = basis.set_input_shape(x)
        >>> basis.n_output_features
        100

        """
        return AtomicBasisMixin.set_input_shape(self, xi)


class CyclicBSplineEval(EvalBasisMixin, CyclicBSplineBasis):
    """
    B-spline 1-dimensional basis functions for cyclic splines.

    Parameters
    ----------
    n_basis_funcs :
        Number of basis functions.
    order :
        Order of the splines used in basis functions. Order must lie within [2, n_basis_funcs].
        The B-splines have (order-2) continuous derivatives at each interior knot.
        The higher this number, the smoother the basis representation will be.
    bounds :
        The bounds for the basis domain. The default ``bounds[0]`` and ``bounds[1]`` are the
        minimum and the maximum of the samples provided when evaluating the basis.
        If a sample is outside the bounds, the basis will return NaN.
    label :
        The label of the basis, intended to be descriptive of the task variable being processed.
        For example: velocity, position, spike_counts.

    Examples
    --------
    >>> from numpy import linspace
    >>> from nemos.basis import CyclicBSplineEval
    >>> n_basis_funcs = 5
    >>> order = 3
    >>> cyclic_bspline_basis = CyclicBSplineEval(n_basis_funcs, order=order)
    >>> cyclic_bspline_basis
    CyclicBSplineEval(n_basis_funcs=5, order=3)
    >>> sample_points = linspace(0, 1, 100)
    >>> features = cyclic_bspline_basis.compute_features(sample_points)
    """

    def __init__(
        self,
        n_basis_funcs: int,
        order: int = 4,
        bounds: Optional[Tuple[float, float]] = None,
        label: Optional[str] = "CyclicBSplineEval",
    ):
        EvalBasisMixin.__init__(self, bounds=bounds)
        CyclicBSplineBasis.__init__(
            self,
            n_basis_funcs,
            mode="eval",
            order=order,
            label=label,
        )

    @add_docstring("split_by_feature", CyclicBSplineBasis)
    def split_by_feature(
        self,
        x: NDArray,
        axis: int = 1,
    ):
        r"""
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import CyclicBSplineEval
        >>> from nemos.glm import GLM
        >>> basis = CyclicBSplineEval(n_basis_funcs=6, label="one_input")
        >>> X = basis.compute_features(np.random.randn(20,))
        >>> split_features_multi = basis.split_by_feature(X, axis=1)
        >>> for feature, sub_dict in split_features_multi.items():
        ...        print(f"{feature}, shape {sub_dict.shape}")
        one_input, shape (20, 6)

        """
        return super().split_by_feature(x, axis=axis)

    @add_docstring("_compute_features", EvalBasisMixin)
    def compute_features(self, xi: ArrayLike) -> FeatureMatrix:
        """
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import CyclicBSplineEval

        >>> # Generate data
        >>> num_samples = 1000
        >>> X = np.random.normal(size=(num_samples, ))  # raw time series
        >>> basis = CyclicBSplineEval(10)
        >>> features = basis.compute_features(X)  # basis transformed time series
        >>> features.shape
        (1000, 10)

        """
        return super().compute_features(xi)

    @add_docstring("evaluate_on_grid", CyclicBSplineBasis)
    def evaluate_on_grid(self, n_samples: int) -> Tuple[NDArray, NDArray]:
        """
        Examples
        --------
        Evaluate and visualize 4 Cyclic B-spline basis functions of order 3:

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from nemos.basis import CyclicBSplineEval
        >>> cbspline_basis = CyclicBSplineEval(n_basis_funcs=4, order=3)
        >>> sample_points, basis_values = cbspline_basis.evaluate_on_grid(100)
        >>> for i in range(4):
        ...     p = plt.plot(sample_points, basis_values[:, i], label=f'Function {i+1}')
        >>> plt.title('Cyclic B-Spline Basis Functions')
        Text(0.5, 1.0, 'Cyclic B-Spline Basis Functions')
        >>> plt.xlabel('Domain')
        Text(0.5, 0, 'Domain')
        >>> plt.ylabel('Basis Function Value')
        Text(0, 0.5, 'Basis Function Value')
        >>> l = plt.legend()
        """
        return super().evaluate_on_grid(n_samples)

    @add_docstring("set_input_shape", AtomicBasisMixin)
    def set_input_shape(self, xi: int | tuple[int, ...] | NDArray):
        """
        Examples
        --------
        >>> import nemos as nmo
        >>> import numpy as np
        >>> basis = nmo.basis.CyclicBSplineEval(5)
        >>> # Configure with an integer input:
        >>> _ = basis.set_input_shape(3)
        >>> basis.n_output_features
        15
        >>> # Configure with a tuple:
        >>> _ = basis.set_input_shape((4, 5))
        >>> basis.n_output_features
        100
        >>> # Configure with an array:
        >>> x = np.ones((10, 4, 5))
        >>> _ = basis.set_input_shape(x)
        >>> basis.n_output_features
        100

        """
        return AtomicBasisMixin.set_input_shape(self, xi)


class CyclicBSplineConv(ConvBasisMixin, CyclicBSplineBasis):
    """
    B-spline 1-dimensional basis functions for cyclic splines.

    Parameters
    ----------
    n_basis_funcs :
        Number of basis functions.
    window_size :
        The window size for convolution in number of samples.
    order :
        Order of the splines used in basis functions. Order must lie within [2, n_basis_funcs].
        The B-splines have (order-2) continuous derivatives at each interior knot.
        The higher this number, the smoother the basis representation will be.
    label :
        The label of the basis, intended to be descriptive of the task variable being processed.
        For example: velocity, position, spike_counts.
    conv_kwargs:
        Additional keyword arguments passed to :func:`nemos.convolve.create_convolutional_predictor`;
        These arguments are used to change the default behavior of the convolution.
        For example, changing the ``predictor_causality``, which by default is set to ``"causal"``.
        Note that one cannot change the default value for the ``axis`` parameter. Basis assumes
        that the convolution axis is ``axis=0``.

    Examples
    --------
    >>> from numpy import linspace
    >>> from nemos.basis import CyclicBSplineConv
    >>> n_basis_funcs = 5
    >>> order = 3
    >>> cyclic_bspline_basis = CyclicBSplineConv(n_basis_funcs, order=order, window_size=10)
    >>> cyclic_bspline_basis
    CyclicBSplineConv(n_basis_funcs=5, window_size=10, order=3)
    >>> sample_points = linspace(0, 1, 100)
    >>> features = cyclic_bspline_basis.compute_features(sample_points)
    """

    def __init__(
        self,
        n_basis_funcs: int,
        window_size: int,
        order: int = 4,
        label: Optional[str] = "CyclicBSplineConv",
        conv_kwargs: Optional[dict] = None,
    ):
        ConvBasisMixin.__init__(self, window_size=window_size, conv_kwargs=conv_kwargs)
        CyclicBSplineBasis.__init__(
            self,
            n_basis_funcs,
            mode="conv",
            order=order,
            label=label,
        )

    @add_docstring("split_by_feature", CyclicBSplineBasis)
    def split_by_feature(
        self,
        x: NDArray,
        axis: int = 1,
    ):
        r"""
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import CyclicBSplineConv
        >>> from nemos.glm import GLM
        >>> basis = CyclicBSplineConv(n_basis_funcs=6, window_size=10, label="two_inputs")
        >>> X_multi = basis.compute_features(np.random.randn(20, 2))
        >>> split_features_multi = basis.split_by_feature(X_multi, axis=1)
        >>> for feature, sub_dict in split_features_multi.items():
        ...        print(f"{feature}, shape {sub_dict.shape}")
        two_inputs, shape (20, 2, 6)

        """
        return super().split_by_feature(x, axis=axis)

    @add_docstring("_compute_features", ConvBasisMixin)
    def compute_features(self, xi: ArrayLike) -> FeatureMatrix:
        """
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import CyclicBSplineConv

        >>> # Generate data
        >>> num_samples = 1000
        >>> X = np.random.normal(size=(num_samples, ))  # raw time series
        >>> basis = CyclicBSplineConv(10, window_size=11)
        >>> features = basis.compute_features(X)  # basis transformed time series
        >>> features.shape
        (1000, 10)

        """
        return super().compute_features(xi)

    @add_docstring("evaluate_on_grid", CyclicBSplineBasis)
    def evaluate_on_grid(self, n_samples: int) -> Tuple[NDArray, NDArray]:
        """
        Examples
        --------
        Evaluate and visualize 4 Cyclic B-spline basis functions of order 3:

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from nemos.basis import CyclicBSplineConv
        >>> cbspline_basis = CyclicBSplineConv(n_basis_funcs=4, order=3, window_size=10)
        >>> sample_points, basis_values = cbspline_basis.evaluate_on_grid(100)
        >>> for i in range(4):
        ...     p = plt.plot(sample_points, basis_values[:, i], label=f'Function {i+1}')
        >>> plt.title('Cyclic B-Spline Basis Functions')
        Text(0.5, 1.0, 'Cyclic B-Spline Basis Functions')
        >>> plt.xlabel('Domain')
        Text(0.5, 0, 'Domain')
        >>> plt.ylabel('Basis Function Value')
        Text(0, 0.5, 'Basis Function Value')
        >>> l = plt.legend()
        """
        return super().evaluate_on_grid(n_samples)

    @add_docstring("set_input_shape", AtomicBasisMixin)
    def set_input_shape(self, xi: int | tuple[int, ...] | NDArray):
        """
        Examples
        --------
        >>> import nemos as nmo
        >>> import numpy as np
        >>> basis = nmo.basis.CyclicBSplineConv(5, 10)
        >>> # Configure with an integer input:
        >>> _ = basis.set_input_shape(3)
        >>> basis.n_output_features
        15
        >>> # Configure with a tuple:
        >>> _ = basis.set_input_shape((4, 5))
        >>> basis.n_output_features
        100
        >>> # Configure with an array:
        >>> x = np.ones((10, 4, 5))
        >>> _ = basis.set_input_shape(x)
        >>> basis.n_output_features
        100

        """
        return AtomicBasisMixin.set_input_shape(self, xi)


class MSplineEval(EvalBasisMixin, MSplineBasis):
    r"""
    M-spline basis functions for modeling and data transformation.

    M-splines [1]_ are a type of spline basis function used for smooth curve fitting
    and data representation. They are positive and integrate to one, making them
    suitable for probabilistic models and density estimation. The order of an
    M-spline defines its smoothness, with higher orders resulting in smoother
    splines.

    This class provides functionality to create M-spline basis functions, allowing
    for flexible and smooth modeling of data. It inherits from the ``SplineBasis``
    abstract class, providing specific implementations for M-splines.

    Parameters
    ----------
    n_basis_funcs :
        The number of basis functions to generate. More basis functions allow for
        more flexible data modeling but can lead to overfitting.
    order :
        The order of the splines used in basis functions. Must be between [1,
        n_basis_funcs]. Default is 2. Higher order splines have more continuous
        derivatives at each interior knot, resulting in smoother basis functions.
    bounds :
        The bounds for the basis domain. The default ``bounds[0]`` and ``bounds[1]`` are the
        minimum and the maximum of the samples provided when evaluating the basis.
        If a sample is outside the bounds, the basis will return NaN.
    label :
        The label of the basis, intended to be descriptive of the task variable being processed.
        For example: velocity, position, spike_counts.

    References
    ----------
    .. [1] Ramsay, J. O. (1988). Monotone regression splines in action. Statistical science,
        3(4), 425-441.

    Notes
    -----
    ``MSplines`` must integrate to 1 over their domain (the area under the curve is 1). Therefore, if the domain
    (x-axis) of an MSpline basis is expanded by a factor of :math:`\alpha`, the values on the co-domain
    (y-axis) values will shrink by a factor of :math:`1/\alpha`.
    For example, over the standard bounds of (0, 1), the maximum value of the MSpline is 18.
    If we set the bounds to (0, 2), the maximum value will be 9, i.e., 18 / 2.

    Examples
    --------
    >>> from numpy import linspace
    >>> from nemos.basis import MSplineEval
    >>> n_basis_funcs = 5
    >>> order = 3
    >>> mspline_basis = MSplineEval(n_basis_funcs, order=order)
    >>> mspline_basis
    MSplineEval(n_basis_funcs=5, order=3)
    >>> sample_points = linspace(0, 1, 100)
    >>> features = mspline_basis.compute_features(sample_points)
    """

    def __init__(
        self,
        n_basis_funcs: int,
        order: int = 4,
        bounds: Optional[Tuple[float, float]] = None,
        label: Optional[str] = "MSplineEval",
    ):
        EvalBasisMixin.__init__(self, bounds=bounds)
        MSplineBasis.__init__(
            self,
            n_basis_funcs,
            mode="eval",
            order=order,
            label=label,
        )

    @add_docstring("split_by_feature", MSplineBasis)
    def split_by_feature(
        self,
        x: NDArray,
        axis: int = 1,
    ):
        r"""
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import MSplineEval
        >>> from nemos.glm import GLM
        >>> basis = MSplineEval(n_basis_funcs=6, label="one_input")
        >>> X = basis.compute_features(np.random.randn(20))
        >>> split_features_multi = basis.split_by_feature(X, axis=1)
        >>> for feature, sub_dict in split_features_multi.items():
        ...        print(f"{feature}, shape {sub_dict.shape}")
        one_input, shape (20, 6)

        """
        return MSplineBasis.split_by_feature(self, x, axis=axis)

    @add_docstring("_compute_features", EvalBasisMixin)
    def compute_features(self, xi: ArrayLike) -> FeatureMatrix:
        """
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import MSplineEval

        >>> # Generate data
        >>> num_samples = 1000
        >>> X = np.random.normal(size=(num_samples, ))  # raw time series
        >>> basis = MSplineEval(10)
        >>> features = basis.compute_features(X)  # basis transformed time series
        >>> features.shape
        (1000, 10)

        """
        return super().compute_features(xi)

    @add_docstring("evaluate_on_grid", MSplineBasis)
    def evaluate_on_grid(self, n_samples: int) -> Tuple[NDArray, NDArray]:
        """
        Examples
        --------
        Evaluate and visualize 4 M-spline basis functions of order 3:

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from nemos.basis import MSplineEval
        >>> mspline_basis = MSplineEval(n_basis_funcs=4, order=3)
        >>> sample_points, basis_values = mspline_basis.evaluate_on_grid(100)
        >>> for i in range(4):
        ...     p = plt.plot(sample_points, basis_values[:, i], label=f'Function {i+1}')
        >>> plt.title('M-Spline Basis Functions')
        Text(0.5, 1.0, 'M-Spline Basis Functions')
        >>> plt.xlabel('Domain')
        Text(0.5, 0, 'Domain')
        >>> plt.ylabel('Basis Function Value')
        Text(0, 0.5, 'Basis Function Value')
        >>> l = plt.legend()
        """
        return super().evaluate_on_grid(n_samples)

    @add_docstring("set_input_shape", AtomicBasisMixin)
    def set_input_shape(self, xi: int | tuple[int, ...] | NDArray):
        """
        Examples
        --------
        >>> import nemos as nmo
        >>> import numpy as np
        >>> basis = nmo.basis.MSplineEval(5)
        >>> # Configure with an integer input:
        >>> _ = basis.set_input_shape(3)
        >>> basis.n_output_features
        15
        >>> # Configure with a tuple:
        >>> _ = basis.set_input_shape((4, 5))
        >>> basis.n_output_features
        100
        >>> # Configure with an array:
        >>> x = np.ones((10, 4, 5))
        >>> _ = basis.set_input_shape(x)
        >>> basis.n_output_features
        100

        """
        return AtomicBasisMixin.set_input_shape(self, xi)


class MSplineConv(ConvBasisMixin, MSplineBasis):
    r"""
    M-spline basis functions for modeling and data transformation.

    M-splines [1]_ are a type of spline basis function used for smooth curve fitting
    and data representation. They are positive and integrate to one, making them
    suitable for probabilistic models and density estimation. The order of an
    M-spline defines its smoothness, with higher orders resulting in smoother
    splines.

    This class provides functionality to create M-spline basis functions, allowing
    for flexible and smooth modeling of data. It inherits from the ``SplineBasis``
    abstract class, providing specific implementations for M-splines.

    Parameters
    ----------
    n_basis_funcs :
        The number of basis functions to generate. More basis functions allow for
        more flexible data modeling but can lead to overfitting.
    order :
        The order of the splines used in basis functions. Must be between [1,
        n_basis_funcs]. Default is 2. Higher order splines have more continuous
        derivatives at each interior knot, resulting in smoother basis functions.
    window_size :
        The window size for convolution in number of samples.
    label :
        The label of the basis, intended to be descriptive of the task variable being processed.
        For example: velocity, position, spike_counts.
    conv_kwargs:
        Additional keyword arguments passed to :func:`nemos.convolve.create_convolutional_predictor`;
        These arguments are used to change the default behavior of the convolution.
        For example, changing the ``predictor_causality``, which by default is set to ``"causal"``.
        Note that one cannot change the default value for the ``axis`` parameter. Basis assumes
        that the convolution axis is ``axis=0``.

    References
    ----------
    .. [1] Ramsay, J. O. (1988). Monotone regression splines in action. Statistical science,
        3(4), 425-441.

    Notes
    -----
    ``MSplines`` must integrate to 1 over their domain (the area under the curve is 1). Therefore, if the domain
    (x-axis) of an MSpline basis is expanded by a factor of :math:`\alpha`, the values on the co-domain
    (y-axis) values will shrink by a factor of :math:`1/\alpha`.
    For example, over the standard bounds of (0, 1), the maximum value of the MSpline is 18.
    If we set the bounds to (0, 2), the maximum value will be 9, i.e., 18 / 2.

    Examples
    --------
    >>> from numpy import linspace
    >>> from nemos.basis import MSplineConv
    >>> n_basis_funcs = 5
    >>> order = 3
    >>> mspline_basis = MSplineConv(n_basis_funcs, order=order, window_size=10)
    >>> mspline_basis
    MSplineConv(n_basis_funcs=5, window_size=10, order=3)
    >>> sample_points = linspace(0, 1, 100)
    >>> features = mspline_basis.compute_features(sample_points)
    """

    def __init__(
        self,
        n_basis_funcs: int,
        window_size: int,
        order: int = 4,
        label: Optional[str] = "MSplineConv",
        conv_kwargs: Optional[dict] = None,
    ):
        ConvBasisMixin.__init__(self, window_size=window_size, conv_kwargs=conv_kwargs)
        MSplineBasis.__init__(
            self,
            n_basis_funcs,
            mode="conv",
            order=order,
            label=label,
        )

    @add_docstring("split_by_feature", MSplineBasis)
    def split_by_feature(
        self,
        x: NDArray,
        axis: int = 1,
    ):
        r"""
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import MSplineConv
        >>> from nemos.glm import GLM
        >>> basis = MSplineConv(n_basis_funcs=6, window_size=10, label="two_inputs")
        >>> X_multi = basis.compute_features(np.random.randn(20, 2))
        >>> split_features_multi = basis.split_by_feature(X_multi, axis=1)
        >>> for feature, sub_dict in split_features_multi.items():
        ...        print(f"{feature}, shape {sub_dict.shape}")
        two_inputs, shape (20, 2, 6)

        """
        return super().split_by_feature(x, axis=axis)

    @add_docstring("_compute_features", ConvBasisMixin)
    def compute_features(self, xi: ArrayLike) -> FeatureMatrix:
        """
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import MSplineConv

        >>> # Generate data
        >>> num_samples = 1000
        >>> X = np.random.normal(size=(num_samples, ))  # raw time series
        >>> basis = MSplineConv(10, window_size=11)
        >>> features = basis.compute_features(X)  # basis transformed time series
        >>> features.shape
        (1000, 10)

        """
        return super().compute_features(xi)

    @add_docstring("evaluate_on_grid", MSplineBasis)
    def evaluate_on_grid(self, n_samples: int) -> Tuple[NDArray, NDArray]:
        """
        Examples
        --------
        Evaluate and visualize 4 M-spline basis functions of order 3:

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from nemos.basis import MSplineConv
        >>> mspline_basis = MSplineConv(n_basis_funcs=4, order=3, window_size=10)
        >>> sample_points, basis_values = mspline_basis.evaluate_on_grid(100)
        >>> for i in range(4):
        ...     p = plt.plot(sample_points, basis_values[:, i], label=f'Function {i+1}')
        >>> plt.title('M-Spline Basis Functions')
        Text(0.5, 1.0, 'M-Spline Basis Functions')
        >>> plt.xlabel('Domain')
        Text(0.5, 0, 'Domain')
        >>> plt.ylabel('Basis Function Value')
        Text(0, 0.5, 'Basis Function Value')
        >>> l = plt.legend()
        """
        return super().evaluate_on_grid(n_samples)

    @add_docstring("set_input_shape", AtomicBasisMixin)
    def set_input_shape(self, xi: int | tuple[int, ...] | NDArray):
        """
        Examples
        --------
        >>> import nemos as nmo
        >>> import numpy as np
        >>> basis = nmo.basis.MSplineConv(5, 10)
        >>> # Configure with an integer input:
        >>> _ = basis.set_input_shape(3)
        >>> basis.n_output_features
        15
        >>> # Configure with a tuple:
        >>> _ = basis.set_input_shape((4, 5))
        >>> basis.n_output_features
        100
        >>> # Configure with an array:
        >>> x = np.ones((10, 4, 5))
        >>> _ = basis.set_input_shape(x)
        >>> basis.n_output_features
        100

        """
        return AtomicBasisMixin.set_input_shape(self, xi)


class RaisedCosineLinearEval(EvalBasisMixin, RaisedCosineBasisLinear):
    """
    Represent linearly-spaced raised cosine basis functions.

    This implementation is based on the cosine bumps used by Pillow et al. [1]_
    to uniformly tile the internal points of the domain.

    Parameters
    ----------
    n_basis_funcs :
        The number of basis functions.
    width :
        Width of the raised cosine. By default, it's set to 2.0.
    bounds :
        The bounds for the basis domain in ``mode="eval"``. The default ``bounds[0]`` and ``bounds[1]`` are the
        minimum and the maximum of the samples provided when evaluating the basis.
        If a sample is outside the bounds, the basis will return NaN.
    label :
        The label of the basis, intended to be descriptive of the task variable being processed.
        For example: velocity, position, spike_counts.

    References
    ----------
    .. [1] Pillow, J. W., Paninski, L., Uzzel, V. J., Simoncelli, E. P., & J.,
        C. E. (2005). Prediction and decoding of retinal ganglion cell responses
        with a probabilistic spiking model. Journal of Neuroscience, 25(47),
        11003–11013.

    Examples
    --------
    >>> import numpy as np
    >>> from nemos.basis import RaisedCosineLinearEval
    >>> n_basis_funcs = 5
    >>> raised_cosine_basis = RaisedCosineLinearEval(n_basis_funcs)
    >>> raised_cosine_basis
    RaisedCosineLinearEval(n_basis_funcs=5, width=2.0)
    >>> sample_points = np.random.randn(100)
    >>> # convolve the basis
    >>> features = raised_cosine_basis.compute_features(sample_points)
    """

    def __init__(
        self,
        n_basis_funcs: int,
        width: float = 2.0,
        bounds: Optional[Tuple[float, float]] = None,
        label: Optional[str] = "RaisedCosineLinearEval",
    ):
        EvalBasisMixin.__init__(self, bounds=bounds)
        RaisedCosineBasisLinear.__init__(
            self,
            n_basis_funcs,
            width=width,
            mode="eval",
            label=label,
        )

    @add_docstring("evaluate_on_grid", RaisedCosineBasisLinear)
    def evaluate_on_grid(self, n_samples: int) -> Tuple[NDArray, NDArray]:
        """
        Examples
        --------
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from nemos.basis import RaisedCosineLinearEval
        >>> n_basis_funcs = 5
        >>> decay_rates = np.array([0.01, 0.02, 0.03, 0.04, 0.05]) # sample decay rates
        >>> window_size=10
        >>> ortho_basis = RaisedCosineLinearEval(n_basis_funcs)
        >>> sample_points, basis_values = ortho_basis.evaluate_on_grid(100)

        """
        return super().evaluate_on_grid(n_samples)

    @add_docstring("_compute_features", EvalBasisMixin)
    def compute_features(self, xi: ArrayLike) -> FeatureMatrix:
        """
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import RaisedCosineLinearEval

        >>> # Generate data
        >>> num_samples = 1000
        >>> X = np.random.normal(size=(num_samples, ))  # raw time series
        >>> basis = RaisedCosineLinearEval(10)
        >>> features = basis.compute_features(X)  # basis transformed time series
        >>> features.shape
        (1000, 10)

        """
        return super().compute_features(xi)

    @add_docstring("split_by_feature", RaisedCosineBasisLinear)
    def split_by_feature(
        self,
        x: NDArray,
        axis: int = 1,
    ):
        r"""
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import RaisedCosineLinearEval
        >>> from nemos.glm import GLM
        >>> basis = RaisedCosineLinearEval(n_basis_funcs=6, label="one_input")
        >>> X = basis.compute_features(np.random.randn(20,))
        >>> split_features_multi = basis.split_by_feature(X, axis=1)
        >>> for feature, sub_dict in split_features_multi.items():
        ...        print(f"{feature}, shape {sub_dict.shape}")
        one_input, shape (20, 6)

        """
        return super().split_by_feature(x, axis=axis)

    @add_docstring("set_input_shape", AtomicBasisMixin)
    def set_input_shape(self, xi: int | tuple[int, ...] | NDArray):
        """
        Examples
        --------
        >>> import nemos as nmo
        >>> import numpy as np
        >>> basis = nmo.basis.RaisedCosineLinearEval(5)
        >>> # Configure with an integer input:
        >>> _ = basis.set_input_shape(3)
        >>> basis.n_output_features
        15
        >>> # Configure with a tuple:
        >>> _ = basis.set_input_shape((4, 5))
        >>> basis.n_output_features
        100
        >>> # Configure with an array:
        >>> x = np.ones((10, 4, 5))
        >>> _ = basis.set_input_shape(x)
        >>> basis.n_output_features
        100

        """
        return AtomicBasisMixin.set_input_shape(self, xi)


class RaisedCosineLinearConv(ConvBasisMixin, RaisedCosineBasisLinear):
    """
    Represent linearly-spaced raised cosine basis functions.

    This implementation is based on the cosine bumps used by Pillow et al. [1]_
    to uniformly tile the internal points of the domain.

    Parameters
    ----------
    n_basis_funcs :
        The number of basis functions.
    width :
        Width of the raised cosine. By default, it's set to 2.0.
    window_size :
        The window size for convolution in number of samples.
    label :
        The label of the basis, intended to be descriptive of the task variable being processed.
        For example: velocity, position, spike_counts.
    conv_kwargs:
        Additional keyword arguments passed to :func:`nemos.convolve.create_convolutional_predictor`;
        These arguments are used to change the default behavior of the convolution.
        For example, changing the ``predictor_causality``, which by default is set to ``"causal"``.
        Note that one cannot change the default value for the ``axis`` parameter. Basis assumes
        that the convolution axis is ``axis=0``.

    References
    ----------
    .. [1] Pillow, J. W., Paninski, L., Uzzel, V. J., Simoncelli, E. P., & J.,
        C. E. (2005). Prediction and decoding of retinal ganglion cell responses
        with a probabilistic spiking model. Journal of Neuroscience, 25(47),
        11003–11013.

    Examples
    --------
    >>> import numpy as np
    >>> from nemos.basis import RaisedCosineLinearConv
    >>> n_basis_funcs = 5
    >>> raised_cosine_basis = RaisedCosineLinearConv(n_basis_funcs, window_size=10)
    >>> raised_cosine_basis
    RaisedCosineLinearConv(n_basis_funcs=5, window_size=10, width=2.0)
    >>> sample_points = np.random.randn(100)
    >>> # convolve the basis
    >>> features = raised_cosine_basis.compute_features(sample_points)
    """

    def __init__(
        self,
        n_basis_funcs: int,
        window_size: int,
        width: float = 2.0,
        label: Optional[str] = "RaisedCosineLinearConv",
        conv_kwargs: Optional[dict] = None,
    ):
        ConvBasisMixin.__init__(self, window_size=window_size, conv_kwargs=conv_kwargs)
        RaisedCosineBasisLinear.__init__(
            self,
            n_basis_funcs,
            mode="conv",
            width=width,
            label=label,
        )

    @add_docstring("evaluate_on_grid", RaisedCosineBasisLinear)
    def evaluate_on_grid(self, n_samples: int) -> Tuple[NDArray, NDArray]:
        """
        Examples
        --------
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from nemos.basis import RaisedCosineLinearConv
        >>> n_basis_funcs = 5
        >>> decay_rates = np.array([0.01, 0.02, 0.03, 0.04, 0.05]) # sample decay rates
        >>> window_size=10
        >>> ortho_basis = RaisedCosineLinearConv(n_basis_funcs, window_size)
        >>> sample_points, basis_values = ortho_basis.evaluate_on_grid(100)

        """
        return super().evaluate_on_grid(n_samples)

    @add_docstring("_compute_features", ConvBasisMixin)
    def compute_features(self, xi: ArrayLike) -> FeatureMatrix:
        """
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import RaisedCosineLinearConv

        >>> # Generate data
        >>> num_samples = 1000
        >>> X = np.random.normal(size=(num_samples, ))  # raw time series
        >>> basis = RaisedCosineLinearConv(10, window_size=100)
        >>> features = basis.compute_features(X)  # basis transformed time series
        >>> features.shape
        (1000, 10)

        """
        return super().compute_features(xi)

    @add_docstring("split_by_feature", RaisedCosineBasisLinear)
    def split_by_feature(
        self,
        x: NDArray,
        axis: int = 1,
    ):
        r"""
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import RaisedCosineLinearConv
        >>> from nemos.glm import GLM
        >>> basis = RaisedCosineLinearConv(n_basis_funcs=6, window_size=10, label="two_inputs")
        >>> X_multi = basis.compute_features(np.random.randn(20, 2))
        >>> split_features_multi = basis.split_by_feature(X_multi, axis=1)
        >>> for feature, sub_dict in split_features_multi.items():
        ...        print(f"{feature}, shape {sub_dict.shape}")
        two_inputs, shape (20, 2, 6)

        """
        return super().split_by_feature(x, axis=axis)

    @add_docstring("set_input_shape", AtomicBasisMixin)
    def set_input_shape(self, xi: int | tuple[int, ...] | NDArray):
        """
        Examples
        --------
        >>> import nemos as nmo
        >>> import numpy as np
        >>> basis = nmo.basis.RaisedCosineLinearConv(5, 10)
        >>> # Configure with an integer input:
        >>> _ = basis.set_input_shape(3)
        >>> basis.n_output_features
        15
        >>> # Configure with a tuple:
        >>> _ = basis.set_input_shape((4, 5))
        >>> basis.n_output_features
        100
        >>> # Configure with an array:
        >>> x = np.ones((10, 4, 5))
        >>> _ = basis.set_input_shape(x)
        >>> basis.n_output_features
        100

        """
        return AtomicBasisMixin.set_input_shape(self, xi)


class RaisedCosineLogEval(EvalBasisMixin, RaisedCosineBasisLog):
    """Represent log-spaced raised cosine basis functions.

    Similar to ``RaisedCosineLinearEval`` but the basis functions are log-spaced.
    This implementation is based on the cosine bumps used by Pillow et al. [1]_
    to uniformly tile the internal points of the domain.

    Parameters
    ----------
    n_basis_funcs :
        The number of basis functions.
    width :
        Width of the raised cosine.
    time_scaling :
        Non-negative hyper-parameter controlling the logarithmic stretch magnitude, with
        larger values resulting in more stretching. As this approaches 0, the
        transformation becomes linear.
    enforce_decay_to_zero:
        If set to True, the algorithm first constructs a basis with ``n_basis_funcs + ceil(width)`` elements
        and subsequently trims off the extra basis elements. This ensures that the final basis element
        decays to 0.
    label :
        The label of the basis, intended to be descriptive of the task variable being processed.
        For example: velocity, position, spike_counts.

    References
    ----------
    .. [1] Pillow, J. W., Paninski, L., Uzzel, V. J., Simoncelli, E. P., & J.,
       C. E. (2005). Prediction and decoding of retinal ganglion cell responses
       with a probabilistic spiking model. Journal of Neuroscience, 25(47),
       11003–11013.

    Examples
    --------
    >>> import numpy as np
    >>> from nemos.basis import RaisedCosineLogEval
    >>> n_basis_funcs = 5
    >>> raised_cosine_basis = RaisedCosineLogEval(n_basis_funcs)
    >>> raised_cosine_basis
    RaisedCosineLogEval(n_basis_funcs=5, width=2.0, time_scaling=50.0, enforce_decay_to_zero=True)
    >>> sample_points = np.random.randn(100)
    >>> # convolve the basis
    >>> features = raised_cosine_basis.compute_features(sample_points)
    """

    def __init__(
        self,
        n_basis_funcs: int,
        width: float = 2.0,
        time_scaling: float = None,
        enforce_decay_to_zero: bool = True,
        bounds: Optional[Tuple[float, float]] = None,
        label: Optional[str] = "RaisedCosineLogEval",
    ):
        EvalBasisMixin.__init__(self, bounds=bounds)
        RaisedCosineBasisLog.__init__(
            self,
            n_basis_funcs,
            width=width,
            time_scaling=time_scaling,
            enforce_decay_to_zero=enforce_decay_to_zero,
            mode="eval",
            label=label,
        )

    @add_docstring("evaluate_on_grid", RaisedCosineBasisLog)
    def evaluate_on_grid(self, n_samples: int) -> Tuple[NDArray, NDArray]:
        """
        Examples
        --------
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from nemos.basis import RaisedCosineLogEval
        >>> n_basis_funcs = 5
        >>> decay_rates = np.array([0.01, 0.02, 0.03, 0.04, 0.05]) # sample decay rates
        >>> window_size=10
        >>> ortho_basis = RaisedCosineLogEval(n_basis_funcs)
        >>> sample_points, basis_values = ortho_basis.evaluate_on_grid(100)

        """
        return super().evaluate_on_grid(n_samples)

    @add_docstring("_compute_features", EvalBasisMixin)
    def compute_features(self, xi: ArrayLike) -> FeatureMatrix:
        """
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import RaisedCosineLogEval

        >>> # Generate data
        >>> num_samples = 1000
        >>> X = np.random.normal(size=(num_samples, ))  # raw time series
        >>> basis = RaisedCosineLogEval(10)
        >>> features = basis.compute_features(X)  # basis transformed time series
        >>> features.shape
        (1000, 10)

        """
        return super().compute_features(xi)

    @add_docstring("split_by_feature", RaisedCosineBasisLog)
    def split_by_feature(
        self,
        x: NDArray,
        axis: int = 1,
    ):
        r"""
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import RaisedCosineLogEval
        >>> from nemos.glm import GLM
        >>> basis = RaisedCosineLogEval(n_basis_funcs=6, label="one_input")
        >>> X = basis.compute_features(np.random.randn(20,))
        >>> split_features_multi = basis.split_by_feature(X, axis=1)
        >>> for feature, sub_dict in split_features_multi.items():
        ...        print(f"{feature}, shape {sub_dict.shape}")
        one_input, shape (20, 6)

        """
        return super().split_by_feature(x, axis=axis)

    @add_docstring("set_input_shape", AtomicBasisMixin)
    def set_input_shape(self, xi: int | tuple[int, ...] | NDArray):
        """
        Examples
        --------
        >>> import nemos as nmo
        >>> import numpy as np
        >>> basis = nmo.basis.RaisedCosineLogEval(5)
        >>> # Configure with an integer input:
        >>> _ = basis.set_input_shape(3)
        >>> basis.n_output_features
        15
        >>> # Configure with a tuple:
        >>> _ = basis.set_input_shape((4, 5))
        >>> basis.n_output_features
        100
        >>> # Configure with an array:
        >>> x = np.ones((10, 4, 5))
        >>> _ = basis.set_input_shape(x)
        >>> basis.n_output_features
        100

        """
        return AtomicBasisMixin.set_input_shape(self, xi)


class RaisedCosineLogConv(ConvBasisMixin, RaisedCosineBasisLog):
    """Represent log-spaced raised cosine basis functions.

    Similar to ``RaisedCosineLinearConv`` but the basis functions are log-spaced.
    This implementation is based on the cosine bumps used by Pillow et al. [1]_
    to uniformly tile the internal points of the domain.

    Parameters
    ----------
    n_basis_funcs :
        The number of basis functions.
    width :
        Width of the raised cosine.
    time_scaling :
        Non-negative hyper-parameter controlling the logarithmic stretch magnitude, with
        larger values resulting in more stretching. As this approaches 0, the
        transformation becomes linear.
    enforce_decay_to_zero:
        If set to True, the algorithm first constructs a basis with ``n_basis_funcs + ceil(width)`` elements
        and subsequently trims off the extra basis elements. This ensures that the final basis element
        decays to 0.
    window_size :
        The window size for convolution. Required if mode is 'conv'.
    label :
        The label of the basis, intended to be descriptive of the task variable being processed.
        For example: velocity, position, spike_counts.
    conv_kwargs:
        Additional keyword arguments passed to :func:`nemos.convolve.create_convolutional_predictor`;
        These arguments are used to change the default behavior of the convolution.
        For example, changing the ``predictor_causality``, which by default is set to ``"causal"``.
        Note that one cannot change the default value for the ``axis`` parameter. Basis assumes
        that the convolution axis is ``axis=0``.

    References
    ----------
    .. [1] Pillow, J. W., Paninski, L., Uzzel, V. J., Simoncelli, E. P., & J.,
       C. E. (2005). Prediction and decoding of retinal ganglion cell responses
       with a probabilistic spiking model. Journal of Neuroscience, 25(47),
       11003–11013.

    Examples
    --------
    >>> import numpy as np
    >>> from nemos.basis import RaisedCosineLogConv
    >>> n_basis_funcs = 5
    >>> raised_cosine_basis = RaisedCosineLogConv(n_basis_funcs, window_size=10)
    >>> raised_cosine_basis
    RaisedCosineLogConv(n_basis_funcs=5, window_size=10, width=2.0, time_scaling=50.0, enforce_decay_to_zero=True)
    >>> sample_points = np.random.randn(100)
    >>> # convolve the basis
    >>> features = raised_cosine_basis.compute_features(sample_points)
    """

    def __init__(
        self,
        n_basis_funcs: int,
        window_size: int,
        width: float = 2.0,
        time_scaling: float = None,
        enforce_decay_to_zero: bool = True,
        label: Optional[str] = "RaisedCosineLogConv",
        conv_kwargs: Optional[dict] = None,
    ):
        ConvBasisMixin.__init__(self, window_size=window_size, conv_kwargs=conv_kwargs)
        RaisedCosineBasisLog.__init__(
            self,
            n_basis_funcs,
            mode="conv",
            width=width,
            time_scaling=time_scaling,
            enforce_decay_to_zero=enforce_decay_to_zero,
            label=label,
        )

    @add_docstring("evaluate_on_grid", RaisedCosineBasisLog)
    def evaluate_on_grid(self, n_samples: int) -> Tuple[NDArray, NDArray]:
        """
        Examples
        --------
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from nemos.basis import RaisedCosineLogConv
        >>> n_basis_funcs = 5
        >>> decay_rates = np.array([0.01, 0.02, 0.03, 0.04, 0.05]) # sample decay rates
        >>> window_size=10
        >>> ortho_basis = RaisedCosineLogConv(n_basis_funcs, window_size)
        >>> sample_points, basis_values = ortho_basis.evaluate_on_grid(100)

        """
        return super().evaluate_on_grid(n_samples)

    @add_docstring("_compute_features", ConvBasisMixin)
    def compute_features(self, xi: ArrayLike) -> FeatureMatrix:
        """
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import RaisedCosineLogConv

        >>> # Generate data
        >>> num_samples = 1000
        >>> X = np.random.normal(size=(num_samples, ))  # raw time series
        >>> basis = RaisedCosineLogConv(10, window_size=100)
        >>> features = basis.compute_features(X)  # basis transformed time series
        >>> features.shape
        (1000, 10)

        """
        return super().compute_features(xi)

    @add_docstring("split_by_feature", RaisedCosineBasisLog)
    def split_by_feature(
        self,
        x: NDArray,
        axis: int = 1,
    ):
        r"""
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import RaisedCosineLogConv
        >>> from nemos.glm import GLM
        >>> basis = RaisedCosineLogConv(n_basis_funcs=6, window_size=10, label="two_inputs")
        >>> X_multi = basis.compute_features(np.random.randn(20, 2))
        >>> split_features_multi = basis.split_by_feature(X_multi, axis=1)
        >>> for feature, sub_dict in split_features_multi.items():
        ...        print(f"{feature}, shape {sub_dict.shape}")
        two_inputs, shape (20, 2, 6)

        """
        return super().split_by_feature(x, axis=axis)

    @add_docstring("set_input_shape", AtomicBasisMixin)
    def set_input_shape(self, xi: int | tuple[int, ...] | NDArray):
        """
        Examples
        --------
        >>> import nemos as nmo
        >>> import numpy as np
        >>> basis = nmo.basis.RaisedCosineLogConv(5, 10)
        >>> # Configure with an integer input:
        >>> _ = basis.set_input_shape(3)
        >>> basis.n_output_features
        15
        >>> # Configure with a tuple:
        >>> _ = basis.set_input_shape((4, 5))
        >>> basis.n_output_features
        100
        >>> # Configure with an array:
        >>> x = np.ones((10, 4, 5))
        >>> _ = basis.set_input_shape(x)
        >>> basis.n_output_features
        100

        """
        return AtomicBasisMixin.set_input_shape(self, xi)


class OrthExponentialEval(EvalBasisMixin, OrthExponentialBasis):
    """Set of 1D basis decaying exponential functions numerically orthogonalized.

    Parameters
    ----------
    n_basis_funcs
        Number of basis functions.
    decay_rates :
        Decay rates of the exponentials, shape ``(n_basis_funcs,)``.
    bounds :
        The bounds for the basis domain. The default ``bounds[0]`` and ``bounds[1]`` are the
        minimum and the maximum of the samples provided when evaluating the basis.
        If a sample is outside the bounds, the basis will return NaN.
    label :
        The label of the basis, intended to be descriptive of the task variable being processed.
        For example: velocity, position, spike_counts.

    Examples
    --------
    >>> import numpy as np
    >>> from numpy import linspace
    >>> from nemos.basis import OrthExponentialEval
    >>> X = np.random.normal(size=(1000, 1))
    >>> n_basis_funcs = 5
    >>> decay_rates = np.array([0.01, 0.02, 0.03, 0.04, 0.05])  # sample decay rates
    >>> window_size = 10
    >>> ortho_basis = OrthExponentialEval(n_basis_funcs, decay_rates)
    >>> ortho_basis
    OrthExponentialEval(n_basis_funcs=5)
    >>> sample_points = linspace(0, 1, 100)
    >>> # evaluate the basis
    >>> features = ortho_basis.compute_features(sample_points)

    """

    def __init__(
        self,
        n_basis_funcs: int,
        decay_rates: NDArray,
        bounds: Optional[Tuple[float, float]] = None,
        label: Optional[str] = "OrthExponentialEval",
    ):
        EvalBasisMixin.__init__(self, bounds=bounds)
        OrthExponentialBasis.__init__(
            self,
            n_basis_funcs,
            decay_rates=decay_rates,
            mode="eval",
            label=label,
        )

    @add_docstring("evaluate_on_grid", OrthExponentialBasis)
    def evaluate_on_grid(self, n_samples: int) -> Tuple[NDArray, NDArray]:
        """
        Examples
        --------
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from nemos.basis import OrthExponentialEval
        >>> n_basis_funcs = 5
        >>> decay_rates = np.array([0.01, 0.02, 0.03, 0.04, 0.05]) # sample decay rates
        >>> window_size=10
        >>> ortho_basis = OrthExponentialEval(n_basis_funcs, decay_rates=decay_rates)
        >>> sample_points, basis_values = ortho_basis.evaluate_on_grid(100)

        """
        return super().evaluate_on_grid(n_samples=n_samples)

    @add_docstring("_compute_features", EvalBasisMixin)
    def compute_features(self, xi: ArrayLike) -> FeatureMatrix:
        """
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import OrthExponentialEval

        >>> # Generate data
        >>> num_samples = 1000
        >>> X = np.random.normal(size=(num_samples, ))  # raw time series
        >>> basis = OrthExponentialEval(10, decay_rates=np.arange(1, 11))
        >>> features = basis.compute_features(X)  # basis transformed time series
        >>> features.shape
        (1000, 10)

        """
        return super().compute_features(xi)

    @add_docstring("split_by_feature", OrthExponentialBasis)
    def split_by_feature(
        self,
        x: NDArray,
        axis: int = 1,
    ):
        """
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import OrthExponentialEval
        >>> from nemos.glm import GLM
        >>> # Define an additive basis
        >>> basis = OrthExponentialEval(n_basis_funcs=5, decay_rates=np.arange(1, 6), label="feature")
        >>> # Generate a sample input array and compute features
        >>> inp = np.random.randn(20)
        >>> X = basis.compute_features(inp)
        >>> # Split the feature matrix along axis 1
        >>> split_features = basis.split_by_feature(X, axis=1)
        >>> for feature, arr in split_features.items():
        ...     print(f"{feature}: shape {arr.shape}")
        feature: shape (20, 5)

        """
        return super().split_by_feature(x, axis=axis)

    @add_docstring("set_input_shape", AtomicBasisMixin)
    def set_input_shape(self, xi: int | tuple[int, ...] | NDArray):
        """
        Examples
        --------
        >>> import nemos as nmo
        >>> import numpy as np
        >>> basis = nmo.basis.OrthExponentialEval(5, decay_rates=np.arange(1, 6))
        >>> # Configure with an integer input:
        >>> _ = basis.set_input_shape(3)
        >>> basis.n_output_features
        15
        >>> # Configure with a tuple:
        >>> _ = basis.set_input_shape((4, 5))
        >>> basis.n_output_features
        100
        >>> # Configure with an array:
        >>> x = np.ones((10, 4, 5))
        >>> _ = basis.set_input_shape(x)
        >>> basis.n_output_features
        100

        """
        return AtomicBasisMixin.set_input_shape(self, xi)


class OrthExponentialConv(ConvBasisMixin, OrthExponentialBasis):
    """Set of 1D basis decaying exponential functions numerically orthogonalized.

    Parameters
    ----------
    n_basis_funcs
        Number of basis functions.
    window_size :
        The window size for convolution in number of samples.
    decay_rates :
        Decay rates of the exponentials, shape ``(n_basis_funcs,)``.
    label :
        The label of the basis, intended to be descriptive of the task variable being processed.
        For example: velocity, position, spike_counts.
    conv_kwargs:
        Additional keyword arguments passed to :func:`nemos.convolve.create_convolutional_predictor`;
        These arguments are used to change the default behavior of the convolution.
        For example, changing the ``predictor_causality``, which by default is set to ``"causal"``.
        Note that one cannot change the default value for the ``axis`` parameter. Basis assumes
        that the convolution axis is ``axis=0``.

    Examples
    --------
    >>> import numpy as np
    >>> from nemos.basis import OrthExponentialConv
    >>> X = np.random.normal(size=(1000, 1))
    >>> n_basis_funcs = 5
    >>> decay_rates = np.array([0.01, 0.02, 0.03, 0.04, 0.05])  # sample decay rates
    >>> window_size = 10
    >>> ortho_basis = OrthExponentialConv(n_basis_funcs, window_size, decay_rates)
    >>> ortho_basis
    OrthExponentialConv(n_basis_funcs=5, window_size=10)
    >>> sample_points = np.random.randn(100)
    >>> # convolve the basis
    >>> features = ortho_basis.compute_features(sample_points)
    """

    def __init__(
        self,
        n_basis_funcs: int,
        window_size: int,
        decay_rates: NDArray,
        label: Optional[str] = "OrthExponentialConv",
        conv_kwargs: Optional[dict] = None,
    ):
        ConvBasisMixin.__init__(self, window_size=window_size, conv_kwargs=conv_kwargs)
        OrthExponentialBasis.__init__(
            self,
            n_basis_funcs,
            mode="conv",
            decay_rates=decay_rates,
            label=label,
        )
        # re-check window size because n_basis_funcs is not set yet when the
        # property setter runs the first check.
        self._check_window_size(self.window_size)

    @add_docstring("evaluate_on_grid", OrthExponentialBasis)
    def evaluate_on_grid(self, n_samples: int) -> Tuple[NDArray, NDArray]:
        """
        Examples
        --------
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from nemos.basis import OrthExponentialConv
        >>> n_basis_funcs = 5
        >>> decay_rates = np.array([0.01, 0.02, 0.03, 0.04, 0.05]) # sample decay rates
        >>> window_size=10
        >>> ortho_basis = OrthExponentialConv(n_basis_funcs, window_size, decay_rates=decay_rates)
        >>> sample_points, basis_values = ortho_basis.evaluate_on_grid(100)

        """
        return super().evaluate_on_grid(n_samples)

    @add_docstring("_compute_features", ConvBasisMixin)
    def compute_features(self, xi: ArrayLike) -> FeatureMatrix:
        """
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import OrthExponentialConv

        >>> # Generate data
        >>> num_samples = 1000
        >>> X = np.random.normal(size=(num_samples, ))  # raw time series
        >>> basis = OrthExponentialConv(10, window_size=100, decay_rates=np.arange(1, 11))
        >>> features = basis.compute_features(X)  # basis transformed time series
        >>> features.shape
        (1000, 10)

        """
        return super().compute_features(xi)

    @add_docstring("split_by_feature", OrthExponentialBasis)
    def split_by_feature(
        self,
        x: NDArray,
        axis: int = 1,
    ):
        r"""
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import OrthExponentialConv
        >>> from nemos.glm import GLM
        >>> basis = OrthExponentialConv(
        ...     n_basis_funcs=6,
        ...     decay_rates=np.arange(1, 7),
        ...     window_size=10,
        ...     label="two_inputs"
        ... )
        >>> X_multi = basis.compute_features(np.random.randn(20, 2))
        >>> split_features_multi = basis.split_by_feature(X_multi, axis=1)
        >>> for feature, sub_dict in split_features_multi.items():
        ...        print(f"{feature}, shape {sub_dict.shape}")
        two_inputs, shape (20, 2, 6)

        """
        return super().split_by_feature(x, axis=axis)

    @add_docstring("set_input_shape", AtomicBasisMixin)
    def set_input_shape(self, xi: int | tuple[int, ...] | NDArray):
        """
        Examples
        --------
        >>> import nemos as nmo
        >>> import numpy as np
        >>> basis = nmo.basis.OrthExponentialConv(5, window_size=10, decay_rates=np.arange(1, 6))
        >>> # Configure with an integer input:
        >>> _ = basis.set_input_shape(3)
        >>> basis.n_output_features
        15
        >>> # Configure with a tuple:
        >>> _ = basis.set_input_shape((4, 5))
        >>> basis.n_output_features
        100
        >>> # Configure with an array:
        >>> x = np.ones((10, 4, 5))
        >>> _ = basis.set_input_shape(x)
        >>> basis.n_output_features
        100

        """
        return AtomicBasisMixin.set_input_shape(self, xi)

    def _check_window_size(self, window_size: int):
        """OrthExponentialBasis specific window size check."""
        super()._check_window_size(window_size)
        # if n_basis_funcs is not yet initialized, skip check
        n_basis = getattr(self, "n_basis_funcs", None)
        if n_basis and window_size < n_basis:
            raise ValueError(
                "OrthExponentialConv basis requires at least a window_size larger then the number "
                f"of basis functions. window_size is {window_size}, n_basis_funcs while"
                f"is {self.n_basis_funcs}."
            )


class IdentityEval(EvalBasisMixin, IdentityBasis):
    """
    Identity basis function.

    This function includes the samples themselves as predictor reshaped as
    a 2D array. It is intended to be used for including a task variable directly as
    a predictor.

    Parameters
    ----------
    bounds :
        The bounds for the basis domain. The default ``bounds[0]`` and ``bounds[1]`` are the
        minimum and the maximum of the samples provided when evaluating the basis.
        If a sample is outside the bounds, the basis will return NaN.
    label :
        The label of the basis, intended to be descriptive of the task variable being processed.
        For example: velocity, position, spike_counts.
    """

    def __init__(
        self,
        bounds: Optional[Tuple[float, float]] = None,
        label: Optional[str] = "IdentityEval",
    ):
        EvalBasisMixin.__init__(self, bounds=bounds)
        IdentityBasis.__init__(
            self,
            n_basis_funcs=1,
            label=label,
        )

    @add_docstring("evaluate_on_grid", OrthExponentialBasis)
    def evaluate_on_grid(self, n_samples: int) -> Tuple[NDArray, NDArray]:
        """
        Examples
        --------
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from nemos.basis import IdentityEval
        >>> basis = IdentityEval()
        >>> sample_points, basis_values = basis.evaluate_on_grid(100)

        """
        return super().evaluate_on_grid(n_samples=n_samples)

    @add_docstring("_compute_features", EvalBasisMixin)
    def compute_features(self, xi: ArrayLike) -> FeatureMatrix:
        """
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import IdentityEval

        >>> # Generate data
        >>> num_samples = 1000
        >>> X = np.random.normal(size=(num_samples, ))  # raw time series
        >>> basis = IdentityEval()
        >>> features = basis.compute_features(X)  # basis transformed time series
        >>> features.shape
        (1000, 1)

        """
        return super().compute_features(xi)

    @add_docstring("split_by_feature", IdentityBasis)
    def split_by_feature(
        self,
        x: NDArray,
        axis: int = 1,
    ):
        """
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import IdentityEval
        >>> from nemos.glm import GLM
        >>> # Define an additive basis
        >>> basis = IdentityEval(label="feature")
        >>> # Generate a sample input array and compute features
        >>> inp = np.random.randn(20)
        >>> X = basis.compute_features(inp)
        >>> # Split the feature matrix along axis 1
        >>> split_features = basis.split_by_feature(X, axis=1)
        >>> for feature, arr in split_features.items():
        ...     print(f"{feature}: shape {arr.shape}")
        feature: shape (20, 1)

        """
        return super().split_by_feature(x, axis=axis)

    @add_docstring("set_input_shape", AtomicBasisMixin)
    def set_input_shape(self, xi: int | tuple[int, ...] | NDArray):
        """
        Examples
        --------
        >>> import nemos as nmo
        >>> import numpy as np
        >>> basis = nmo.basis.IdentityEval()
        >>> # Configure with an integer input:
        >>> _ = basis.set_input_shape(3)
        >>> basis.n_output_features
        3
        >>> # Configure with a tuple:
        >>> _ = basis.set_input_shape((4, 5))
        >>> basis.n_output_features
        20
        >>> # Configure with an array:
        >>> x = np.ones((10, 4, 5))
        >>> _ = basis.set_input_shape(x)
        >>> basis.n_output_features
        20

        """
        return AtomicBasisMixin.set_input_shape(self, xi)


class HistoryConv(ConvBasisMixin, HistoryBasis):
    """Basis for history effects.

    This function includes the history of the samples as predictor reshaped as
    a 2D array. It is intended to be used for including a raw history as predictor.

    Parameters
    ----------
    window_size:
        History window as the number of samples.
    label :
        The label of the basis, intended to be descriptive of the task variable being processed.
        For example: velocity, position, spike_counts.
    conv_kwargs:
        Additional keyword arguments passed to :func:`nemos.convolve.create_convolutional_predictor`;
        These arguments are used to change the default behavior of the convolution.
        For example, changing the ``predictor_causality``, which by default is set to ``"causal"``.
        Note that one cannot change the default value for the ``axis`` parameter. Basis assumes
        that the convolution axis is ``axis=0``.
    """

    def __init__(
        self,
        window_size: int,
        label: Optional[str] = "HistoryConv",
        conv_kwargs: Optional[dict] = None,
    ):
        ConvBasisMixin.__init__(self, window_size=window_size, conv_kwargs=conv_kwargs)
        HistoryBasis.__init__(
            self,
            n_basis_funcs=window_size,
            label=label,
        )

    @add_docstring("evaluate_on_grid", HistoryBasis)
    def evaluate_on_grid(self, n_samples: int) -> Tuple[NDArray, NDArray]:
        """
        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> from nemos.basis import HistoryConv
        >>> window_size=100
        >>> basis = HistoryConv(window_size=window_size)
        >>> sample_points, basis_values = basis.evaluate_on_grid(window_size)

        """
        return super().evaluate_on_grid(n_samples)

    @add_docstring("_compute_features", ConvBasisMixin)
    def compute_features(self, xi: ArrayLike) -> FeatureMatrix:
        """
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import HistoryConv

        >>> # Generate data
        >>> num_samples = 1000
        >>> X = np.random.normal(size=(num_samples, ))  # raw time series
        >>> basis = HistoryConv(10)
        >>> features = basis.compute_features(X)  # basis transformed time series
        >>> features.shape
        (1000, 10)

        """
        return super().compute_features(xi)

    @add_docstring("split_by_feature", IdentityBasis)
    def split_by_feature(
        self,
        x: NDArray,
        axis: int = 1,
    ):
        """
        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import HistoryConv
        >>> from nemos.glm import GLM
        >>> # Define an additive basis
        >>> basis = HistoryConv(5, label="feature")
        >>> # Generate a sample input array and compute features
        >>> inp = np.random.randn(20)
        >>> X = basis.compute_features(inp)
        >>> # Split the feature matrix along axis 1
        >>> split_features = basis.split_by_feature(X, axis=1)
        >>> for feature, arr in split_features.items():
        ...     print(f"{feature}: shape {arr.shape}")
        feature: shape (20, 5)

        """
        return super().split_by_feature(x, axis=axis)

    @add_docstring("set_input_shape", AtomicBasisMixin)
    def set_input_shape(self, xi: int | tuple[int, ...] | NDArray):
        """
        Examples
        --------
        >>> import nemos as nmo
        >>> import numpy as np
        >>> basis = nmo.basis.HistoryConv(5)
        >>> # Configure with an integer input:
        >>> _ = basis.set_input_shape(3)
        >>> basis.n_output_features
        15
        >>> # Configure with a tuple:
        >>> _ = basis.set_input_shape((4, 5))
        >>> basis.n_output_features
        100
        >>> # Configure with an array:
        >>> x = np.ones((10, 4, 5))
        >>> _ = basis.set_input_shape(x)
        >>> basis.n_output_features
        100

        """
        return AtomicBasisMixin.set_input_shape(self, xi)

    @property
    def window_size(self):
        return self._window_size

    @window_size.setter
    def window_size(self, window_size: int) -> None:
        self._check_window_size(window_size)
        self._window_size = window_size
        self._n_basis_funcs = window_size



"""
Module for fetching data and utilities using the pooch library.

This module allows you to download datasets and utility scripts
from specified URLs and manage their versions and integrity using
SHA256 hash verification. It also provides helper functions to
calculate hashes for local files and manage file paths.
"""

import os
import pathlib
from typing import Optional, Union

try:
    import pooch
    from pooch import Pooch
    from tqdm.auto import tqdm
except ImportError:
    pooch = None
    Pooch = None
    tqdm = None

try:
    import dandi
    import fsspec
    import h5py
    from dandi.dandiapi import DandiAPIClient
    from fsspec.implementations.cached import CachingFileSystem
    from pynwb import NWBHDF5IO
except ImportError:
    dandi = None
    NWBHDF5IO = None


# Registry of dataset filenames and their corresponding SHA256 hashes.
REGISTRY_DATA = {
    "A0670-221213.nwb": "8587dd6dde107504bd4a17a68ce8fb934fcbcccc337e653f31484611ee51f50a",
    "Mouse32-140822.nwb": "1a919a033305b8f58b5c3e217577256183b14ed5b436d9c70989dee6dafe0f35",
    "Achilles_10252013.nwb": "42857015aad4c2f7f6f3d4022611a69bc86d714cf465183ce30955731e614990",
    "allen_478498617.nwb": "262393d7485a5b39cc80fb55011dcf21f86133f13d088e35439c2559fd4b49fa",
    "m691l1.nwb": "1990d8d95a70a29af95dade51e60ffae7a176f6207e80dbf9ccefaf418fe22b6",
    "A2929-200711.nwb": "f698d7319efa5dfeb18fb5fe718ec1a84fdf96b85a158177849a759cd5e396fe",
}
DOWNLOADABLE_FILES = list(REGISTRY_DATA.keys())

# URL templates for downloading datasets and utility scripts.
OSF_TEMPLATE = "https://osf.io/{}/download"

# Mapping of dataset filenames to their download URLs.
REGISTRY_URLS_DATA = {
    "A0670-221213.nwb": OSF_TEMPLATE.format("sbnaw"),
    "Mouse32-140822.nwb": OSF_TEMPLATE.format("jb2gd"),
    "Achilles_10252013.nwb": OSF_TEMPLATE.format("hu5ma"),
    "allen_478498617.nwb": OSF_TEMPLATE.format("vf2nj"),
    "m691l1.nwb": OSF_TEMPLATE.format("xesdm"),
    "A2929-200711.nwb": OSF_TEMPLATE.format("y7zwd"),
}

_NEMOS_ENV = "NEMOS_DATA_DIR"


def _create_retriever(path: Optional[pathlib.Path] = None) -> Pooch:
    """Create a pooch retriever for fetching datasets.

    This function sets up the pooch retriever, which manages the
    downloading and caching of files, including handling retries
    and checking file integrity using SHA256 hashes.

    Parameters
    ----------
    path :
        The directory where datasets will be stored. If not provided,
        defaults to pooch's cache (check ``pooch.os_cache('nemos')`` for that path)

    Returns
    -------
    :
        A configured pooch retriever object.

    """
    if path is None:
        # Use the default data directory if none is provided.
        path = pooch.os_cache("nemos")

    return pooch.create(
        path=path,
        base_url="",
        urls=REGISTRY_URLS_DATA,
        registry=REGISTRY_DATA,
        retry_if_failed=2,
        allow_updates="POOCH_ALLOW_UPDATES",
        env=_NEMOS_ENV,
    )


def fetch_data(
    dataset_name: str, path: Optional[Union[pathlib.Path, str]] = None
) -> str:
    """
    Download a dataset using pooch.

    This function downloads a dataset, checking if it already exists
    and is unchanged. If the dataset is an archive (ends in .tar.gz),
    it decompresses the archive and returns the path to the resulting
    directory. Otherwise, it returns the path to the downloaded file.

    Parameters
    ----------
    dataset_name :
        The name of the dataset to download. Must match an entry in
        REGISTRY_DATA.
    path :
        The directory where the dataset will be stored. If not provided,
        defaults to _DEFAULT_DATA_DIR.

    Returns
    -------
    :
        The path to the downloaded file or directory.
    """
    if pooch is None:
        raise ImportError(
            "Missing optional dependency 'pooch'."
            " Please use pip or "
            "conda to install 'pooch'."
        )
    retriever = _create_retriever(path)
    # Fetch the dataset using pooch.
    return retriever.fetch(dataset_name)


def download_dandi_data(dandiset_id: str, filepath: str) -> NWBHDF5IO:
    """Download a dataset from the DANDI Archive (https://dandiarchive.org/)

    Parameters
    ----------
    dandiset_id :
        6-character string of numbers giving the ID of the dandiset.
    filepath :
        filepath to the specific .nwb file within the dandiset we wish to return.

    Returns
    -------
    io :
        NWB file containing specified data.

    Examples
    --------
    >>> import nemos as nmo
    >>> import pynapple as nap
    >>> io = nmo.fetch.download_dandi_data("000582",
    ...                                    "sub-11265/sub-11265_ses-07020602_behavior+ecephys.nwb")
    >>> nwb = nap.NWBFile(io.read(), lazy_loading=False)
    >>> print(nwb)
    07020602
    ┍━━━━━━━━━━━━━━━━━━━━━┯━━━━━━━━━━┑
    │ Keys                │ Type     │
    ┝━━━━━━━━━━━━━━━━━━━━━┿━━━━━━━━━━┥
    │ units               │ TsGroup  │
    │ ElectricalSeriesLFP │ Tsd      │
    │ SpatialSeriesLED2   │ TsdFrame │
    │ SpatialSeriesLED1   │ TsdFrame │
    │ ElectricalSeries    │ Tsd      │
    ┕━━━━━━━━━━━━━━━━━━━━━┷━━━━━━━━━━┙

    """
    if dandi is None:
        raise ImportError(
            "Missing optional dependency 'dandi'."
            " Please use pip or "
            "conda to install 'dandi'."
        )
    with DandiAPIClient() as client:
        asset = client.get_dandiset(dandiset_id, "draft").get_asset_by_path(filepath)
        s3_url = asset.get_content_url(follow_redirects=1, strip_query=True)

    # first, create a virtual filesystem based on the http protocol
    fs = fsspec.filesystem("http")

    # create a cache to save downloaded data to disk (optional)
    # mimicking caching behavior of pooch create
    if _NEMOS_ENV in os.environ:
        cache_dir = pathlib.Path(os.environ[_NEMOS_ENV])
    else:
        cache_dir = pooch.os_cache("nemos") / "nwb-cache"
    cache_dir.mkdir(parents=True, exist_ok=True)

    fs = CachingFileSystem(
        fs=fs,
        cache_storage=cache_dir.as_posix(),  # Local folder for the cache
    )

    # next, open the file
    file = h5py.File(fs.open(s3_url, "rb"))
    io = NWBHDF5IO(file=file, load_namespaces=True)

    return io



"""Decaying exponential basis."""

# required to get ArrayLike to render correctly
from __future__ import annotations

import abc
from typing import Optional, Tuple

import numpy as np
import scipy.linalg
from numpy.typing import ArrayLike, NDArray
from pynapple import Tsd, TsdFrame, TsdTensor

from ..type_casting import support_pynapple
from ..typing import FeatureMatrix
from ._basis import Basis, check_transform_input, min_max_rescale_samples
from ._basis_mixin import AtomicBasisMixin


class OrthExponentialBasis(Basis, AtomicBasisMixin, abc.ABC):
    """Set of 1D basis decaying exponential functions numerically orthogonalized.

    Parameters
    ----------
    n_basis_funcs
            Number of basis functions.
    decay_rates :
            Decay rates of the exponentials, shape ``(n_basis_funcs,)``.
    mode :
        The mode of operation. ``'eval'`` for evaluation at sample points,
        ``'conv'`` for convolutional operation.
    label :
        The label of the basis, intended to be descriptive of the task variable being processed.
        For example: velocity, position, spike_counts.
    """

    def __init__(
        self,
        n_basis_funcs: int,
        decay_rates: NDArray[np.floating],
        mode="eval",
        label: Optional[str] = "OrthExponentialBasis",
    ):
        AtomicBasisMixin.__init__(self, n_basis_funcs=n_basis_funcs)
        super().__init__(
            mode=mode,
            label=label,
        )
        self.decay_rates = decay_rates
        self._check_rates()
        self._n_input_dimensionality = 1

    @property
    def decay_rates(self):
        r"""Decay rate.

        The rate of decay of the exponential functions. If :math:`f_i(t) = e^{-\alpha_i t}` is the i-th decay
        exponential before orthogonalization, :math:`\alpha_i` is the i-th element of the ``decay_rate`` vector.
        """
        return self._decay_rates

    @decay_rates.setter
    def decay_rates(self, value: NDArray):
        """Decay rate setter."""
        value = np.asarray(value, dtype=float)
        if value.shape[0] != self.n_basis_funcs:
            raise ValueError(
                f"The number of basis functions must match the number of decay rates provided. "
                f"Number of basis functions provided: {self.n_basis_funcs}, "
                f"Number of decay rates provided: {value.shape[0]}"
            )
        self._decay_rates = value

    def _check_n_basis_min(self) -> None:
        """Check that the user required enough basis elements.

        Checks that the number of basis is at least 1.

        Raises
        ------
        ValueError
            If an insufficient number of basis element is requested for the basis type
        """
        if self.n_basis_funcs < 1:
            raise ValueError(
                f"Object class {self.__class__.__name__} requires >= 1 basis elements. "
                f"{self.n_basis_funcs} basis elements specified instead"
            )

    def _check_rates(self) -> None:
        """
        Check if the decay rates list has duplicate entries.

        Raises
        ------
        ValueError
            If two or more decay rates are repeated, which would result in a linearly
            dependent set of functions for the basis.
        """
        if len(set(self._decay_rates)) != len(self._decay_rates):
            raise ValueError(
                "Two or more rates are repeated! Repeating rates will result in a "
                "linearly dependent set of function for the basis."
            )

    def _check_sample_size(self, *sample_pts: NDArray) -> None:
        """Check that the sample size is greater than the number of basis.

        This is necessary for the orthogonalization procedure,
        that otherwise will return (sample_size, ) basis elements instead of the expected number.

        Parameters
        ----------
        sample_pts
            Spacing for basis functions, holding elements on the interval [0, inf).

        Raises
        ------
        ValueError
            If the number of basis element is less than the number of samples.
        """
        if sample_pts[0].size < self.n_basis_funcs:
            raise ValueError(
                "OrthExponentialBasis requires at least as many samples as basis functions!\n"
                f"Class instantiated with {self.n_basis_funcs} basis functions "
                f"but only {sample_pts[0].size} samples provided!"
            )

    @support_pynapple(conv_type="numpy")
    @check_transform_input
    def _evaluate(
        self,
        sample_pts: ArrayLike | Tsd | TsdFrame | TsdTensor,
    ) -> FeatureMatrix:
        """Generate basis functions with given spacing.

        Parameters
        ----------
        sample_pts
            Spacing for basis functions, holding elements on the interval :math:`[0,inf)`.
            `sample_pts` is a n-dimensional (n >= 1) array with first axis being the samples, i.e.
            `sample_pts.shape[0] == n_samples`.

        Returns
        -------
        basis_funcs
            Evaluated exponentially decaying basis functions, numerically
            orthogonalized, shape ``(n_samples, n_basis_funcs)``.

        """
        self._check_sample_size(sample_pts)
        sample_pts, _ = min_max_rescale_samples(
            sample_pts, getattr(self, "bounds", None)
        )

        # process one input at the time (orthogonalization must be done one input at the time)

        # flatten all non-time dimensions
        shape = sample_pts.shape
        sample_pts = sample_pts.reshape(sample_pts.shape[0], -1)
        basis_list = []
        max_rank = 0
        for samp in sample_pts.T:
            valid_idx = ~np.isnan(samp)
            # because of how scipy.linalg.orth works, have to create a matrix of
            # shape (n_pts, n_basis_funcs) and then transpose, rather than
            # directly computing orth on the matrix of shape (n_basis_funcs,
            # n_pts)
            exp_decay_eval = np.stack(
                [np.exp(-lam * samp[valid_idx]) for lam in self._decay_rates], axis=1
            )
            # count the linear independent components (could be lower than n_basis_funcs for num precision).
            n_independent_component = np.linalg.matrix_rank(exp_decay_eval)
            # initialize output to nan
            basis_funcs = np.full(
                shape=(samp.shape[0], n_independent_component), fill_value=np.nan
            )
            # orthonormalize on valid points
            basis_funcs[valid_idx] = scipy.linalg.orth(exp_decay_eval)

            max_rank = max(max_rank, basis_funcs.shape[1])

            # append the basis
            basis_list.append(basis_funcs)

        # fill the value in
        out = np.full(
            shape=(shape[0], len(basis_list), max_rank),
            fill_value=np.nan,
            dtype=np.float64,
        )
        for i, bas in enumerate(basis_list):
            out[:, i, : bas.shape[1]] = bas

        # reshape back
        out = out.reshape(*shape, -1)
        return out

    def evaluate_on_grid(self, n_samples: int) -> Tuple[Tuple[NDArray], NDArray]:
        """Evaluate the basis set on a grid of equi-spaced sample points.

        Parameters
        ----------
        n_samples :
            The number of points in the uniformly spaced grid. A higher number of
            samples will result in a more detailed visualization of the basis functions.

        Returns
        -------
        X :
            Array of shape (n_samples,) containing the equi-spaced sample
            points where we've evaluated the basis.
        basis_funcs :
            OrthExponential basis functions, shape (n_samples, n_basis_funcs).
        """
        return super().evaluate_on_grid(n_samples)



"""
A Module for Optimization with Various Regularization Schemes.

This module provides a series of classes that facilitate the optimization of models
with different types of regularizations. Each `Regularizer` class in this module interfaces
with various optimization methods, and they can be applied depending on the model's requirements.
"""

import abc
from typing import Callable, Tuple, Union

import jax
import jax.numpy as jnp
import jaxopt
from numpy.typing import NDArray

from . import tree_utils
from .base_class import Base
from .proximal_operator import prox_group_lasso
from .typing import DESIGN_INPUT_TYPE, ProximalOperator
from .utils import format_repr

__all__ = ["UnRegularized", "Ridge", "Lasso", "GroupLasso"]


def __dir__() -> list[str]:
    return __all__


class Regularizer(Base, abc.ABC):
    """
    Abstract base class for regularized solvers.

    This class is designed to provide a consistent interface for optimization solvers,
    enabling users to easily switch between different regularizers, ensuring compatibility
    with various loss functions and optimization algorithms.

    Attributes
    ----------
    allowed_solvers :
        Tuple of solver names that are allowed for use with this regularizer.
    default_solver :
        String of the default solver name allowed for use with this regularizer.
    """

    _allowed_solvers: Tuple[str] = tuple()
    _default_solver: str = None

    def __init__(
        self,
        **kwargs,
    ):
        super().__init__(**kwargs)

    @property
    def allowed_solvers(self) -> Tuple[str]:
        return self._allowed_solvers

    @property
    def default_solver(self) -> str:
        return self._default_solver

    @abc.abstractmethod
    def penalized_loss(self, loss: Callable, regularizer_strength: float) -> Callable:
        """
        Abstract method to penalize loss functions.

        Parameters
        ----------
        loss :
            Callable loss function.
        regularizer_strength :
            Float the indicates the regularization strength.

        Returns
        -------
        :
            A modified version of the loss function including any relevant penalization based on the regularizer
            type.
        """
        pass

    @abc.abstractmethod
    def get_proximal_operator(
        self,
    ) -> ProximalOperator:
        """
        Abstract method to retrieve the proximal operator for this solver.

        Returns
        -------
        :
            The proximal operator, which typically applies a form of regularization.
        """
        pass

    def __repr__(self):
        return format_repr(self)


class UnRegularized(Regularizer):
    """
    Regularizer class for unregularized models.

    This class equips models with the identity proximal operator (no shrinkage) and the
    unpenalized loss function.
    """

    _allowed_solvers = (
        "GradientDescent",
        "BFGS",
        "LBFGS",
        "NonlinearCG",
        "ProximalGradient",
        "SVRG",
        "ProxSVRG",
    )

    _default_solver = "GradientDescent"

    def __init__(
        self,
    ):
        super().__init__()

    def penalized_loss(self, loss: Callable, regularizer_strength: float):
        """
        Returns the original loss function unpenalized.

        Unregularized regularization method does not add any penalty.
        """
        return loss

    def get_proximal_operator(
        self,
    ) -> ProximalOperator:
        """
        Returns the identity operator.

        Unregularized method corresponds to an identity proximal operator, since no
        shrinkage factor is applied.
        """

        return jaxopt.prox.prox_none


class Ridge(Regularizer):
    """
    Regularizer class for Ridge (L2 regularization).

    This class equips models with the Ridge proximal operator and the
    Ridge penalized loss function.
    """

    _allowed_solvers = (
        "GradientDescent",
        "BFGS",
        "LBFGS",
        "NonlinearCG",
        "ProximalGradient",
        "SVRG",
        "ProxSVRG",
    )

    _default_solver = "GradientDescent"

    def __init__(
        self,
    ):
        super().__init__()

    @staticmethod
    def _penalization(
        params: Tuple[DESIGN_INPUT_TYPE, jnp.ndarray], regularizer_strength: float
    ) -> jnp.ndarray:
        """
        Compute the Ridge penalization for given parameters.

        Parameters
        ----------
        params :
            Model parameters for which to compute the penalization.

        Returns
        -------
        float
            The Ridge penalization value.
        """

        def l2_penalty(coeff: jnp.ndarray, intercept: jnp.ndarray) -> jnp.ndarray:
            return (
                0.5
                * regularizer_strength
                * jnp.sum(jnp.power(coeff, 2))
                / intercept.shape[0]
            )

        # tree map the computation and sum over leaves
        return tree_utils.pytree_map_and_reduce(
            lambda x: l2_penalty(x, params[1]), sum, params[0]
        )

    def penalized_loss(self, loss: Callable, regularizer_strength: float) -> Callable:
        """Returns the penalized loss function for Ridge regularization."""

        def _penalized_loss(params, X, y):
            return loss(params, X, y) + self._penalization(params, regularizer_strength)

        return _penalized_loss

    def get_proximal_operator(
        self,
    ) -> ProximalOperator:
        """
        Retrieve the proximal operator for Ridge regularization (L2 penalty).

        Returns
        -------
        :
            The proximal operator, applying L2 regularization to the provided parameters. The intercept
            term is not regularized.
        """

        def prox_op(params, l2reg, scaling=1.0):
            Ws, bs = params
            l2reg /= bs.shape[0]
            return jaxopt.prox.prox_ridge(Ws, l2reg, scaling=scaling), bs

        return prox_op


class Lasso(Regularizer):
    """
    Regularizer class for Lasso (L1 regularization).

    This class equips models with the Lasso proximal operator and the
    Lasso penalized loss function.
    """

    _allowed_solvers = (
        "ProximalGradient",
        "ProxSVRG",
    )

    _default_solver = "ProximalGradient"

    def __init__(
        self,
    ):
        super().__init__()

    def get_proximal_operator(
        self,
    ) -> ProximalOperator:
        """
        Retrieve the proximal operator for Lasso regularization (L1 penalty).

        Returns
        -------
        :
            The proximal operator, applying L1 regularization to the provided parameters. The intercept
            term is not regularized.
        """

        def prox_op(params, l1reg, scaling=1.0):
            Ws, bs = params
            l1reg /= bs.shape[0]
            # if Ws is a pytree, l1reg needs to be a pytree with the same
            # structure
            l1reg = jax.tree_util.tree_map(lambda x: l1reg * jnp.ones_like(x), Ws)
            return jaxopt.prox.prox_lasso(Ws, l1reg, scaling=scaling), bs

        return prox_op

    @staticmethod
    def _penalization(
        params: Tuple[DESIGN_INPUT_TYPE, jnp.ndarray], regularizer_strength: float
    ) -> jnp.ndarray:
        """
        Compute the Lasso penalization for given parameters.

        Parameters
        ----------
        params :
            Model parameters for which to compute the penalization.

        Returns
        -------
        float
            The Lasso penalization value.
        """

        def l1_penalty(coeff: jnp.ndarray, intercept: jnp.ndarray) -> jnp.ndarray:
            return regularizer_strength * jnp.sum(jnp.abs(coeff)) / intercept.shape[0]

        # tree map the computation and sum over leaves
        return tree_utils.pytree_map_and_reduce(
            lambda x: l1_penalty(x, params[1]), sum, params[0]
        )

    def penalized_loss(self, loss: Callable, regularizer_strength: float) -> Callable:
        """Returns a function for calculating the penalized loss using Lasso regularization."""

        def _penalized_loss(params, X, y):
            return loss(params, X, y) + self._penalization(params, regularizer_strength)

        return _penalized_loss


class GroupLasso(Regularizer):
    """
    Regularizer class for Group Lasso (group-L1) regularized models.

    This class equips models with the group-lasso proximal operator and the
    group-lasso penalized loss function.

    Attributes
    ----------
    mask :
        A 2d mask array indicating groups of features for regularization, shape ``(num_groups, num_features)``.
        Each row represents a group of features.
        Each column corresponds to a feature, where a value of 1 indicates that the feature belongs
        to the group, and a value of 0 indicates it doesn't.
        Default is ``mask = np.ones((1, num_features))``, grouping all features in a single group.

    Examples
    --------
    >>> import numpy as np
    >>> from nemos.regularizer import GroupLasso  # Assuming the module is named group_lasso
    >>> from nemos.glm import GLM
    >>> # simulate some counts
    >>> num_samples, num_features, num_groups = 1000, 5, 3
    >>> X = np.random.normal(size=(num_samples, num_features)) # design matrix
    >>> w = [0, 0.5, 1, 0, -0.5] # define some weights
    >>> y = np.random.poisson(np.exp(X.dot(w))) # observed counts
    >>> # Define a mask for 3 groups and 5 features
    >>> mask = np.zeros((num_groups, num_features))
    >>> mask[0] = [1, 0, 0, 1, 0]  # Group 0 includes features 0 and 3
    >>> mask[1] = [0, 1, 0, 0, 0]  # Group 1 includes features 1
    >>> mask[2] = [0, 0, 1, 0, 1]  # Group 2 includes features 2 and 4

    >>> # Create the GroupLasso regularizer instance
    >>> group_lasso = GroupLasso(mask=mask)
    >>> # fit a group-lasso glm
    >>> model = GLM(regularizer=group_lasso, regularizer_strength=0.1).fit(X, y)
    >>> print(f"coeff shape: {model.coef_.shape}")
    coeff shape: (5,)
    """

    _allowed_solvers = (
        "ProximalGradient",
        "ProxSVRG",
    )

    _default_solver = "ProximalGradient"

    def __init__(
        self,
        mask: Union[NDArray, jnp.ndarray] = None,
    ):
        super().__init__()

        self.mask = mask

    @property
    def mask(self):
        """Getter for the mask attribute."""
        return self._mask

    @mask.setter
    def mask(self, mask: Union[jnp.ndarray, None]):
        """Setter for the mask attribute."""
        # check mask if passed by user, else will be initialized later
        if mask is not None:
            self._check_mask(mask)
        self._mask = mask

    @staticmethod
    def _check_mask(mask: jnp.ndarray):
        """
        Validate the mask array.

        This method ensures the mask adheres to requirements:
        - It should be 2-dimensional.
        - Each element must be either 0 or 1.
        - Each feature should belong to only one group.
        - The mask should not be empty.
        - The mask is an array of float type.

        Raises
        ------
        ValueError
            If any of the above conditions are not met.
        """
        if mask.ndim != 2:
            raise ValueError(
                "`mask` must be 2-dimensional. "
                f"{mask.ndim} dimensional mask provided instead!"
            )

        if mask.shape[0] == 0:
            raise ValueError(f"Empty mask provided! Mask has shape {mask.shape}.")

        if jnp.any((mask != 1) & (mask != 0)):
            raise ValueError("Mask elements be 0s and 1s!")

        if mask.sum() == 0:
            raise ValueError("Empty mask provided!")

        if jnp.any(mask.sum(axis=0) > 1):
            raise ValueError(
                "Incorrect group assignment. Some of the features are assigned "
                "to more than one group."
            )

        if not jnp.issubdtype(mask.dtype, jnp.floating):
            raise ValueError(
                "Mask should be a floating point jnp.ndarray. "
                f"Data type {mask.dtype} provided instead!"
            )

    def _penalization(
        self, params: Tuple[DESIGN_INPUT_TYPE, jnp.ndarray], regularizer_strength: float
    ) -> jnp.ndarray:
        r"""
        Calculate the penalization.

        Note: the penalty is being calculated according to the following formula:

        .. math::

            \\text{loss}(\beta_1,...,\beta_g) + \alpha \cdot \sum _{j=1...,g} \sqrt{\dim(\beta_j)} || \beta_j||_2

        where :math:`g` is the number of groups, :math:`\dim(\cdot)` is the dimension of the vector,
        i.e. the number of coefficient in each :math:`\beta_j`, and :math:`||\cdot||_2` is the euclidean norm.
        """

        # conform to shape (1, n_features) if param is (n_features,) or (n_neurons, n_features) if
        # param is (n_features, n_neurons)
        param_with_extra_axis = jnp.atleast_2d(params[0].T)

        vec_prod = jax.vmap(
            lambda x: self.mask * x, in_axes=0, out_axes=2
        )  # this vectorizes the product over the neurons, and adds the neuron axis as the last axis

        masked_param = vec_prod(
            param_with_extra_axis
        )  # this masks the param, (group, feature, neuron)

        penalty = jax.numpy.sum(
            jax.numpy.linalg.norm(masked_param, axis=1).T
            * jax.numpy.sqrt(self.mask.sum(axis=1))
        )

        # divide regularization strength by number of neurons
        regularizer_strength = regularizer_strength / params[1].shape[0]

        return penalty * regularizer_strength

    def penalized_loss(self, loss: Callable, regularizer_strength: float) -> Callable:
        """Returns a function for calculating the penalized loss using Group Lasso regularization."""

        def _penalized_loss(params, X, y):
            return loss(params, X, y) + self._penalization(params, regularizer_strength)

        return _penalized_loss

    def get_proximal_operator(
        self,
    ) -> ProximalOperator:
        """
        Retrieve the proximal operator for Group Lasso regularization.

        Returns
        -------
        :
            The proximal operator, applying Group Lasso regularization to the provided parameters. The
            intercept term is not regularized.
        """

        def prox_op(params, regularizer_strength, scaling=1.0):
            return prox_group_lasso(
                params, regularizer_strength, mask=self.mask, scaling=scaling
            )

        return prox_op



#!/usr/bin/env python3
__version__ = "0.2.1"

from . import (
    basis,
    convolve,
    exceptions,
    fetch,
    glm,
    identifiability_constraints,
    observation_models,
    pytrees,
    regularizer,
    simulation,
    styles,
    tree_utils,
    type_casting,
    utils,
)



#!/usr/bin/env python3

from ._myst_nb_glue import FormattedString, glue_two_step_convolve
from .plotting import (
    PlotSlidingWindow,
    current_injection_plot,
    highlight_max_cell,
    lnp_schematic,
    plot_basis,
    plot_convolved_counts,
    plot_count_history_window,
    plot_coupling,
    plot_features,
    plot_head_direction_tuning,
    plot_head_direction_tuning_model,
    plot_heatmap_cv_results,
    plot_history_window,
    plot_position_phase_speed_tuning,
    plot_position_speed_tuning,
    plot_rates_and_smoothed_counts,
    plot_weighted_sum_basis,
    run_animation,
    tuning_curve_plot,
)
